"""
YOLO Engine - Fast multi-node generation with minimal questions

Generates multiple agents and their flow in one go based on a single requirement.
"""
import os
from pathlib import Path
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

from mofa.vibe.llm_client import LLMClient
from mofa.vibe.engine import VibeEngine
from mofa.vibe.models import VibeConfig, AgentCode
from mofa.vibe.scaffolder import ProjectScaffolder


class YoloEngine:
    """YOLO mode engine for fast multi-node generation"""

    def __init__(self, requirement: str, llm_model: str, api_key: str,
                 agents_output: str, flows_output: str):
        self.requirement = requirement
        self.llm_model = llm_model
        self.api_key = api_key
        self.agents_output = agents_output
        self.flows_output = flows_output
        self.console = Console()
        self.llm = LLMClient(model=llm_model, api_key=api_key)

    def run(self):
        """Run YOLO generation"""
        try:
            # Step 1: Analyze requirement and generate plan
            self.console.print("\n[cyan]Analyzing requirement...[/cyan]")
            plan = self._generate_plan()

            # Step 2: Generate new agents and collect reused agents
            total_agents = len(plan.get('reuse_agents', [])) + len(plan.get('new_agents', []))
            self.console.print(f"\n[cyan]Processing {total_agents} agents...[/cyan]")
            agent_paths = self._generate_agents(plan)

            # Step 3: Generate flow
            self.console.print("\n[cyan]Generating flow...[/cyan]")
            flow_path = self._generate_flow(plan, agent_paths)

            # Step 4: Auto-run flow test
            self.console.print("\n[cyan]Running flow test...[/cyan]")
            self._run_flow_test(flow_path)

            # Collect all agent names (reused + new)
            all_agents = []
            if 'reuse_agents' in plan:
                all_agents.extend([a['name'] for a in plan['reuse_agents']])
            if 'new_agents' in plan:
                all_agents.extend([a['name'] for a in plan['new_agents']])

            return {
                'flow_path': flow_path,
                'agents': all_agents
            }

        except Exception as e:
            self.console.print(f"[red]Error: {e}[/red]")
            raise

    def _scan_existing_agents(self):
        """Scan existing agents directory and extract agent information"""
        agents_info = []
        agents_dir = Path(self.agents_output)

        if not agents_dir.exists():
            return agents_info

        # Iterate through agent directories
        for agent_dir in agents_dir.iterdir():
            # Skip hidden files and non-directories
            if agent_dir.name.startswith('.') or not agent_dir.is_dir():
                continue

            # Skip terminal-input (special agent)
            if agent_dir.name == 'terminal-input':
                continue

            try:
                agent_info = self._extract_agent_info(agent_dir)
                if agent_info:
                    agents_info.append(agent_info)
            except Exception as e:
                # Skip agents that can't be parsed
                continue

        return agents_info

    def _extract_agent_info(self, agent_dir):
        """Extract information from a single agent"""
        agent_name = agent_dir.name
        main_py = agent_dir / 'agent' / 'main.py'
        readme = agent_dir / 'README.md'

        if not main_py.exists():
            return None

        # Read main.py
        with open(main_py, 'r', encoding='utf-8') as f:
            code = f.read()

        # Extract input/output parameters
        import re
        inputs = re.findall(r"receive_parameter\(['\"](\w+)['\"]\)", code)
        outputs = re.findall(r"agent_output_name=['\"](\w+)['\"]", code)

        # Read description from README if exists
        description = ""
        if readme.exists():
            with open(readme, 'r', encoding='utf-8') as f:
                readme_content = f.read()
                # Try to extract first meaningful line
                lines = [l.strip() for l in readme_content.split('\n') if l.strip() and not l.startswith('#')]
                if lines and not lines[0].startswith('Auto-generated'):
                    description = lines[0]

        # If no description, try to infer from code comments
        if not description:
            comment_match = re.search(r'#\s*(.+)', code)
            if comment_match:
                description = comment_match.group(1).strip()

        # Fallback: generate description from agent name
        if not description:
            description = f"Agent: {agent_name.replace('-', ' ')}"

        return {
            'name': agent_name,
            'description': description[:100],  # Limit length
            'inputs': inputs,
            'outputs': outputs,
            'code_snippet': code[:300]  # First 300 chars for context
        }

    def _generate_plan(self):
        """Use LLM to analyze requirement and create generation plan"""
        # Scan existing agents
        existing_agents = self._scan_existing_agents()

        # Build context about existing agents
        existing_context = ""
        if existing_agents:
            existing_context = "\n\nExisting agents you can reuse:\n"
            for agent in existing_agents:
                existing_context += f"- {agent['name']}: {agent['description']}\n"
                existing_context += f"  Input: {', '.join(agent['inputs']) if agent['inputs'] else 'none'}\n"
                existing_context += f"  Output: {', '.join(agent['outputs']) if agent['outputs'] else 'none'}\n"

        prompt = f"""Analyze this requirement and create a plan for generating MoFA agents and their dataflow:
{existing_context}

Requirement: {self.requirement}

You need to:
1. Check if existing agents can be reused for the required functionality
2. Only create NEW agents for functionality that doesn't exist
3. Break down into 2-6 agents maximum
4. Design the dataflow connections - can be linear, parallel, or any topology

Return a JSON object with this structure:
{{
    "reuse_agents": [
        {{
            "name": "existing-agent-name",
            "reason": "Why reusing this agent"
        }}
    ],
    "new_agents": [
        {{
            "name": "new-agent-name",
            "description": "What this agent does",
            "input_param": "parameter_name",
            "output_param": "result_name"
        }}
    ],
    "connections": [
        {{
            "from": "source-agent-name (or 'terminal-input')",
            "from_param": "output_parameter_name (or 'data' for terminal-input)",
            "to": "target-agent-name",
            "to_param": "input_parameter_name"
        }}
    ]
}}

IMPORTANT RULES:
1. Prefer reusing existing agents when their functionality matches
2. Define ALL connections explicitly in the "connections" array
3. The dataflow must start from "terminal-input" with output "data"
4. The dataflow must end by sending back to "terminal-input" with input "agent_response"
5. You can have parallel branches, multiple inputs, or any topology as needed
6. Make sure input/output parameter names match the actual agent definitions

Example connection patterns:
- Linear: terminal-input -> agent1 -> agent2 -> terminal-input
- Parallel: terminal-input -> [agent1, agent2] -> merger -> terminal-input
- Multi-input: terminal-input -> agent1 -> agent3
                            └-> agent2 -> agent3
"""

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            console=self.console,
            transient=True
        ) as progress:
            task = progress.add_task("Planning...", total=None)

            # Use LLMClient's generate method
            response_text = self.llm.generate(prompt)

            import json
            plan_text = response_text.strip()

            # Extract JSON from markdown code blocks if present
            if "```json" in plan_text:
                plan_text = plan_text.split("```json")[1].split("```")[0].strip()
            elif "```" in plan_text:
                plan_text = plan_text.split("```")[1].split("```")[0].strip()

            plan = json.loads(plan_text)

        # Display plan
        self.console.print(f"\n[bold]Plan:[/bold]")

        if 'reuse_agents' in plan and plan['reuse_agents']:
            self.console.print(f"\n[cyan]Reusing existing agents:[/cyan]")
            for agent in plan['reuse_agents']:
                self.console.print(f"  ✓ {agent['name']}: {agent.get('reason', '')}")

        if 'new_agents' in plan and plan['new_agents']:
            self.console.print(f"\n[yellow]Creating new agents:[/yellow]")
            for i, agent in enumerate(plan['new_agents'], 1):
                self.console.print(f"  {i}. {agent['name']}: {agent['description']}")

        return plan

    def _generate_agent_code_yolo(self, agent_name: str, description: str, input_param: str, output_param: str) -> str:
        """Generate agent code in YOLO mode (without test cases)"""

        # Read LLM config from .env
        import os
        from dotenv import load_dotenv
        load_dotenv()

        api_key = os.getenv('OPENAI_API_KEY', '')
        base_url = os.getenv('OPENAI_API_BASE', 'https://api.openai.com/v1')
        model = os.getenv('MOFA_VIBE_MODEL', 'gpt-4o-mini')

        # Build LLM integration example with actual values
        llm_example = ""
        if api_key:
            llm_example = f"""

## LLM Integration (if needed)
If the description involves LLM/AI calls, you can use this pre-configured OpenAI client:

```python
from openai import OpenAI

# Pre-configured with your .env settings
client = OpenAI(
    api_key="{api_key}",
    base_url="{base_url}"
)

response = client.chat.completions.create(
    model="{model}",
    messages=[
        {{"role": "system", "content": "You are a helpful assistant."}},
        {{"role": "user", "content": user_input}}
    ],
    temperature=0.7
)

result = response.choices[0].message.content
```

Note: The API key, base URL, and model are from your .env configuration.
"""

        prompt = f"""
Generate a complete MoFA agent implementation for the following requirement.

## Agent Name
{agent_name}

## Description
{description}

## Input/Output Parameters
- Input Parameter: {input_param}
- Output Parameter: {output_param}

## MoFA Agent Template
You MUST follow this exact structure:

```python
from mofa.agent_build.base.base_agent import MofaAgent, run_agent

@run_agent
def run(agent: MofaAgent):
    # Step 1: Receive input parameter
    {input_param} = agent.receive_parameter('{input_param}')

    # Step 2: Implement the business logic based on the description
    # Your code here to process the input

    # Step 3: Send output
    agent.send_output(agent_output_name='{output_param}', agent_result=result)

def main():
    agent = MofaAgent(agent_name='{agent_name}')
    run(agent=agent)

if __name__ == "__main__":
    main()
```

## Guidelines
1. Import necessary libraries at the top
2. Follow Python best practices
3. Add error handling if necessary
4. Keep it simple and focused on the description
5. The input parameter name is '{input_param}', output parameter name is '{output_param}'
6. Output ONLY the complete Python code, no explanations or markdown
{llm_example}
Generate the complete main.py code now:
"""

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{{task.description}}"),
            console=self.console,
            transient=True
        ) as progress:
            task = progress.add_task(f"Generating {agent_name}...", total=None)

            code_str = self.llm.generate(prompt)

            # Clean up code (remove markdown blocks)
            if "```python" in code_str:
                code_str = code_str.split("```python")[1].split("```")[0].strip()
            elif "```" in code_str:
                code_str = code_str.split("```")[1].split("```")[0].strip()

        return code_str

    def _generate_flow_name(self, agent_names):
        """Generate a meaningful flow name based on agent names"""
        agents_str = ", ".join(agent_names)

        prompt = f"""Generate a concise, descriptive flow name for a dataflow with these agents:

Agents: {agents_str}
Original requirement: {self.requirement}

Guidelines:
1. Use lowercase letters, numbers, and hyphens only
2. 2-4 words maximum
3. Focus on what the FLOW does (not individual agents)
4. Use simple, clear English words
5. Examples:
   - agents: [text-processor, sentiment-analyzer] → "text-sentiment-analysis"
   - agents: [image-loader, face-detector, age-estimator] → "face-age-detection"
   - agents: [data-fetcher, csv-parser, stats-calculator] → "data-statistics-pipeline"

Output ONLY the flow name (e.g., "text-analyzer-flow"), no explanations or quotes:
"""

        name = self.llm.generate(prompt).strip().strip('"').strip("'")

        # Ensure it's valid (only lowercase, numbers, hyphens)
        import re
        name = re.sub(r'[^a-z0-9-]', '-', name.lower())
        name = re.sub(r'-+', '-', name)  # Remove duplicate hyphens
        name = name.strip('-')  # Remove leading/trailing hyphens

        # Fallback if empty or too short
        if len(name) < 3:
            name = f"flow-{len(agent_names)}-agents"

        return name

    def _generate_agents(self, plan):
        """Generate new agents and collect paths of reused agents"""
        agent_paths = {}

        # First, add paths for reused agents
        if 'reuse_agents' in plan:
            for agent_spec in plan['reuse_agents']:
                agent_name = agent_spec['name']
                agent_path = Path(self.agents_output) / agent_name
                if agent_path.exists():
                    agent_paths[agent_name] = str(agent_path)
                    self.console.print(f"  [green]✓[/green] {agent_name} (reused)")

        # Then generate new agents
        if 'new_agents' in plan:
            for agent_spec in plan['new_agents']:
                agent_name = agent_spec['name']
                agent_desc = agent_spec['description']
                input_param = agent_spec.get('input_param', 'data')
                output_param = agent_spec.get('output_param', 'result')

                self.console.print(f"\n  Generating {agent_name}...")

                # Generate code directly using LLM (no tests needed in YOLO mode)
                code_str = self._generate_agent_code_yolo(agent_name, agent_desc, input_param, output_param)

                # Create project using scaffolder
                from mofa.vibe.scaffolder import ProjectScaffolder

                scaffolder = ProjectScaffolder(output_dir=self.agents_output)
                # In YOLO mode, we don't need tests, pass empty test_yaml
                agent_path = scaffolder.create_project(
                    agent_name=agent_name,
                    code=code_str,
                    test_yaml="",  # No tests in YOLO mode
                )
                agent_paths[agent_name] = agent_path

                self.console.print(f"  [green]✓[/green] {agent_name}")

        return agent_paths

    def _generate_flow(self, plan, agent_paths):
        """Generate flow YAML connecting all agents based on explicit connections"""
        import yaml

        # Collect all agents info
        all_agents = {}

        # Add reused agents with their existing info
        if 'reuse_agents' in plan:
            for agent_spec in plan['reuse_agents']:
                agent_info = self._get_agent_io_from_existing(agent_spec['name'])
                if agent_info:
                    all_agents[agent_spec['name']] = {
                        'input_param': agent_info['input_param'],
                        'output_param': agent_info['output_param']
                    }

        # Add new agents
        if 'new_agents' in plan:
            for agent_spec in plan['new_agents']:
                all_agents[agent_spec['name']] = {
                    'input_param': agent_spec['input_param'],
                    'output_param': agent_spec['output_param']
                }

        # Build connection map: {agent_name: {input_param: source}}
        connections = plan.get('connections', [])
        agent_inputs = {}  # {agent_name: {param_name: source}}
        terminal_input_source = None  # What feeds back to terminal-input

        for conn in connections:
            from_agent = conn['from']
            from_param = conn['from_param']
            to_agent = conn['to']
            to_param = conn['to_param']

            source = f"{from_agent}/{from_param}"

            if to_agent == 'terminal-input':
                # This is the feedback to terminal-input
                terminal_input_source = source
            else:
                # Regular agent connection
                if to_agent not in agent_inputs:
                    agent_inputs[to_agent] = {}
                agent_inputs[to_agent][to_param] = source

        # If no explicit terminal-input feedback, use last agent
        if not terminal_input_source and all_agents:
            last_agent_name = list(all_agents.keys())[-1]
            last_agent_output = all_agents[last_agent_name]['output_param']
            terminal_input_source = f"{last_agent_name}/{last_agent_output}"

        # Generate meaningful flow name using LLM
        flow_name = self._generate_flow_name(list(all_agents.keys()))
        flow_dir = Path(self.flows_output) / flow_name
        flow_dir.mkdir(parents=True, exist_ok=True)

        # Create nodes list
        nodes = []

        # Add terminal-input as first node
        nodes.append({
            'id': 'terminal-input',
            'build': f'pip install -e ../../agents/terminal-input',
            'path': 'dynamic',
            'outputs': ['data'],
            'inputs': {
                'agent_response': terminal_input_source
            }
        })

        # Add all other agents with their connections
        for agent_name, agent_info in all_agents.items():
            output_param = agent_info['output_param']

            # Get inputs for this agent
            inputs = agent_inputs.get(agent_name, {})

            # If no explicit inputs, default to terminal-input/data
            if not inputs:
                input_param = agent_info['input_param']
                inputs = {input_param: 'terminal-input/data'}

            node = {
                'id': agent_name,
                'build': f'pip install -e ../../agents/{agent_name}',
                'path': agent_name,
                'outputs': [output_param],
                'inputs': inputs
            }

            # Check if this is the final agent (feeds back to terminal-input)
            if terminal_input_source and terminal_input_source.startswith(f"{agent_name}/"):
                node['env'] = {
                    'IS_DATAFLOW_END': True,
                    'WRITE_LOG': True
                }

            nodes.append(node)

        flow_data = {'nodes': nodes}

        # Write flow YAML
        flow_yaml_path = flow_dir / f'{flow_name}_dataflow.yml'
        with open(flow_yaml_path, 'w') as f:
            yaml.dump(flow_data, f, default_flow_style=False, sort_keys=False)

        return str(flow_yaml_path)

    def _get_agent_io_from_existing(self, agent_name):
        """Extract input/output params from existing agent"""
        agent_dir = Path(self.agents_output) / agent_name
        main_py = agent_dir / 'agent' / 'main.py'

        if not main_py.exists():
            return None

        try:
            with open(main_py, 'r', encoding='utf-8') as f:
                code = f.read()

            import re
            inputs = re.findall(r"receive_parameter\(['\"](\w+)['\"]\)", code)
            outputs = re.findall(r"agent_output_name=['\"](\w+)['\"]", code)

            if inputs and outputs:
                return {
                    'input_param': inputs[0],
                    'output_param': outputs[0]
                }
        except Exception:
            pass

        return None

    def _run_flow_test(self, flow_yaml_path):
        """Auto-run flow test once with better interrupt handling"""
        import subprocess
        import signal

        self.console.print("\n" + "=" * 60)
        self.console.print("[bold cyan]Auto Flow Testing[/bold cyan]")
        self.console.print("=" * 60)
        self.console.print("Testing the flow with terminal-input...")
        self.console.print("[dim]Tip: Type 'exit' in terminal-input to quit gracefully[/dim]")
        self.console.print("[dim]Or press Ctrl+C to force stop[/dim]")
        self.console.print("=" * 60 + "\n")

        process = None
        try:
            # Run the flow test
            process = subprocess.Popen(
                ['mofa', 'run-flow', flow_yaml_path],
                cwd=str(Path(flow_yaml_path).parent),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1
            )

            # Stream output in real-time
            for line in process.stdout:
                print(line, end='')

            # Wait for process to complete
            process.wait()

        except KeyboardInterrupt:
            self.console.print("\n\n[yellow]Interrupted by user (Ctrl+C)[/yellow]")

            # Gracefully terminate the process
            if process and process.poll() is None:
                try:
                    # Try graceful termination first
                    process.terminate()
                    process.wait(timeout=3)
                except subprocess.TimeoutExpired:
                    # Force kill if it doesn't terminate
                    process.kill()
                    process.wait()

                self.console.print("[dim]Flow process terminated[/dim]")

        except Exception as e:
            self.console.print(f"\n[red]Error during flow test: {e}[/red]")
            if process and process.poll() is None:
                process.kill()
                process.wait()



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import List

from .char_convert import tranditional_to_simplified
from .chronology import RE_DATE
from .chronology import RE_DATE2
from .chronology import RE_TIME
from .chronology import RE_TIME_RANGE
from .chronology import replace_date
from .chronology import replace_date2
from .chronology import replace_time
from .constants import F2H_ASCII_LETTERS
from .constants import F2H_DIGITS
from .constants import F2H_SPACE
from .num import RE_DECIMAL_NUM
from .num import RE_DEFAULT_NUM
from .num import RE_FRAC
from .num import RE_INTEGER
from .num import RE_NUMBER
from .num import RE_PERCENTAGE
from .num import RE_POSITIVE_QUANTIFIERS
from .num import RE_RANGE
from .num import RE_TO_RANGE
from .num import RE_ASMD
from .num import RE_POWER
from .num import replace_default_num
from .num import replace_frac
from .num import replace_negative_num
from .num import replace_number
from .num import replace_percentage
from .num import replace_positive_quantifier
from .num import replace_range
from .num import replace_to_range
from .num import replace_asmd
from .num import replace_power
from .phonecode import RE_MOBILE_PHONE
from .phonecode import RE_NATIONAL_UNIFORM_NUMBER
from .phonecode import RE_TELEPHONE
from .phonecode import replace_mobile
from .phonecode import replace_phone
from .quantifier import RE_TEMPERATURE
from .quantifier import replace_measure
from .quantifier import replace_temperature


class TextNormalizer():
    def __init__(self):
        self.SENTENCE_SPLITOR = re.compile(r'([：、，；。？！,;?!][”’]?)')

    def _split(self, text: str, lang="zh") -> List[str]:
        """Split long text into sentences with sentence-splitting punctuations.
        Args:
            text (str): The input text.
        Returns:
            List[str]: Sentences.
        """
        # Only for pure Chinese here
        if lang == "zh":
            text = text.replace(" ", "")
            # 过滤掉特殊字符
            text = re.sub(r'[——《》【】<>{}()（）#&@“”^_|\\]', '', text)
        text = self.SENTENCE_SPLITOR.sub(r'\1\n', text)
        text = text.strip()
        sentences = [sentence.strip() for sentence in re.split(r'\n+', text)]
        return sentences

    def _post_replace(self, sentence: str) -> str:
        sentence = sentence.replace('/', '每')
        # sentence = sentence.replace('~', '至')
        # sentence = sentence.replace('～', '至')
        sentence = sentence.replace('①', '一')
        sentence = sentence.replace('②', '二')
        sentence = sentence.replace('③', '三')
        sentence = sentence.replace('④', '四')
        sentence = sentence.replace('⑤', '五')
        sentence = sentence.replace('⑥', '六')
        sentence = sentence.replace('⑦', '七')
        sentence = sentence.replace('⑧', '八')
        sentence = sentence.replace('⑨', '九')
        sentence = sentence.replace('⑩', '十')
        sentence = sentence.replace('α', '阿尔法')
        sentence = sentence.replace('β', '贝塔')
        sentence = sentence.replace('γ', '伽玛').replace('Γ', '伽玛')
        sentence = sentence.replace('δ', '德尔塔').replace('Δ', '德尔塔')
        sentence = sentence.replace('ε', '艾普西龙')
        sentence = sentence.replace('ζ', '捷塔')
        sentence = sentence.replace('η', '依塔')
        sentence = sentence.replace('θ', '西塔').replace('Θ', '西塔')
        sentence = sentence.replace('ι', '艾欧塔')
        sentence = sentence.replace('κ', '喀帕')
        sentence = sentence.replace('λ', '拉姆达').replace('Λ', '拉姆达')
        sentence = sentence.replace('μ', '缪')
        sentence = sentence.replace('ν', '拗')
        sentence = sentence.replace('ξ', '克西').replace('Ξ', '克西')
        sentence = sentence.replace('ο', '欧米克伦')
        sentence = sentence.replace('π', '派').replace('Π', '派')
        sentence = sentence.replace('ρ', '肉')
        sentence = sentence.replace('ς', '西格玛').replace('Σ', '西格玛').replace(
            'σ', '西格玛')
        sentence = sentence.replace('τ', '套')
        sentence = sentence.replace('υ', '宇普西龙')
        sentence = sentence.replace('φ', '服艾').replace('Φ', '服艾')
        sentence = sentence.replace('χ', '器')
        sentence = sentence.replace('ψ', '普赛').replace('Ψ', '普赛')
        sentence = sentence.replace('ω', '欧米伽').replace('Ω', '欧米伽')
        # 兜底数学运算，顺便兼容懒人用语
        sentence = sentence.replace('+', '加')
        sentence = sentence.replace('-', '减')
        sentence = sentence.replace('×', '乘')
        sentence = sentence.replace('÷', '除')
        sentence = sentence.replace('=', '等')
        # re filter special characters, have one more character "-" than line 68
        sentence = re.sub(r'[-——《》【】<=>{}()（）#&@“”^_|\\]', '', sentence)
        return sentence

    def normalize_sentence(self, sentence: str) -> str:
        # basic character conversions
        sentence = tranditional_to_simplified(sentence)
        sentence = sentence.translate(F2H_ASCII_LETTERS).translate(
            F2H_DIGITS).translate(F2H_SPACE)

        # number related NSW verbalization
        sentence = RE_DATE.sub(replace_date, sentence)
        sentence = RE_DATE2.sub(replace_date2, sentence)

        # range first
        sentence = RE_TIME_RANGE.sub(replace_time, sentence)
        sentence = RE_TIME.sub(replace_time, sentence)

        # 处理~波浪号作为至的替换
        sentence = RE_TO_RANGE.sub(replace_to_range, sentence)
        sentence = RE_TEMPERATURE.sub(replace_temperature, sentence)
        sentence = replace_measure(sentence)

        # 处理数学运算
        while RE_ASMD.search(sentence):
            sentence = RE_ASMD.sub(replace_asmd, sentence)
        sentence = RE_POWER.sub(replace_power, sentence)

        sentence = RE_FRAC.sub(replace_frac, sentence)
        sentence = RE_PERCENTAGE.sub(replace_percentage, sentence)
        sentence = RE_MOBILE_PHONE.sub(replace_mobile, sentence)

        sentence = RE_TELEPHONE.sub(replace_phone, sentence)
        sentence = RE_NATIONAL_UNIFORM_NUMBER.sub(replace_phone, sentence)

        sentence = RE_RANGE.sub(replace_range, sentence)

        sentence = RE_INTEGER.sub(replace_negative_num, sentence)
        sentence = RE_DECIMAL_NUM.sub(replace_number, sentence)
        sentence = RE_POSITIVE_QUANTIFIERS.sub(replace_positive_quantifier,
                                               sentence)
        sentence = RE_DEFAULT_NUM.sub(replace_default_num, sentence)
        sentence = RE_NUMBER.sub(replace_number, sentence)
        sentence = self._post_replace(sentence)

        return sentence

    def normalize(self, text: str) -> List[str]:
        sentences = self._split(text)
        sentences = [self.normalize_sentence(sent) for sent in sentences]
        return sentences



"""
This file bundles language identification functions.

Modifications (fork): Copyright (c) 2021, Adrien Barbaresi.

Original code: Copyright (c) 2011 Marco Lui <saffsd@gmail.com>.
Based on research by Marco Lui and Tim Baldwin.

See LICENSE file for more info.
https://github.com/adbar/py3langid

Projects:
https://github.com/juntaosun/LangSegment
"""

import re
from collections import defaultdict

# import langid
import py3langid as langid
# pip install py3langid==0.2.2


# -----------------------------------
# 更新日志：新版本分词更加精准。
# Changelog: The new version of the word segmentation is more accurate.
# -----------------------------------


# Word segmentation function:
# automatically identify and split the words (Chinese/English/Japanese/Korean) in the article or sentence according to different languages,
# making it more suitable for TTS processing.
# This code is designed for front-end text multi-lingual mixed annotation distinction, multi-language mixed training and inference of various TTS projects.
# This processing result is mainly for (Chinese = zh, Japanese = ja, English = en, Korean = ko), and can actually support up to 97 different language mixing processing.


# ===========================================================================================================
# 分词功能：将文章或句子里的例如（中/英/日/韩），按不同语言自动识别并拆分，让它更适合TTS处理。
# 本代码专为各种 TTS 项目的前端文本多语种混合标注区分，多语言混合训练和推理而编写。
# ===========================================================================================================
# （1）自动分词：“韩语中的오빠读什么呢？あなたの体育の先生は誰ですか? 此次发布会带来了四款iPhone 15系列机型”
# （2）手动分词：“你的名字叫<ja>佐々木？<ja>吗？”
# 本处理结果主要针对（中文=zh , 日文=ja , 英文=en , 韩语=ko）, 实际上可支持多达 97 种不同的语言混合处理。
# ===========================================================================================================


# 手动分词标签规范：<语言标签>文本内容</语言标签>
# Manual word segmentation tag specification: <language tags> text content </language tags>
# ===========================================================================================================
# For manual word segmentation, labels need to appear in pairs, such as:
# 如需手动分词，标签需要成对出现，例如：“<ja>佐々木<ja>”  或者  “<ja>佐々木</ja>”
# 错误示范：“你的名字叫<ja>佐々木。” 此句子中出现的单个<ja>标签将被忽略，不会处理。
# Error demonstration: "Your name is <ja>佐々木。" Single <ja> tags that appear in this sentence will be ignored and will not be processed.
# ===========================================================================================================

class LangSegment():

    _text_cache = None
    _text_lasts = None
    _text_langs = None
    _lang_count = None
    _lang_eos =   None

    # 可自定义语言匹配标签：
    # Customizable language matching tags: These are supported
    # <zh>你好<zh> , <ja>佐々木</ja> , <en>OK<en> , <ko>오빠</ko> 这些写法均支持
    SYMBOLS_PATTERN = r'(<([a-zA-Z|-]*)>(.*?)<\/*[a-zA-Z|-]*>)'

    # 语言过滤组功能, 可以指定保留语言。不在过滤组中的语言将被清除。您可随心搭配TTS语音合成所支持的语言。
    # The language filter group function allows you to specify reserved languages.
    # Languages not in the filter group will be cleared. You can match the languages supported by TTS Text To Speech as you like.
    Langfilters = ["zh", "en", "ja", "ko"]
    # 除此以外，它支持简写过滤器，只需按不同语种任意组合即可。
    # In addition to that, it supports abbreviation filters, allowing for any combination of different languages.
    # 示例：您可以任意指定多种组合，进行过滤
    # Example: You can specify any combination to filter

    # Langfilters = ["zh"]              # 按中文识别
    # Langfilters = ["en"]              # 按英文识别
    # Langfilters = ["ja"]              # 按日文识别
    # Langfilters = ["ko"]              # 按韩文识别
    # Langfilters = ["zh_ja"]           # 中日混合识别
    # Langfilters = ["zh_en"]           # 中英混合识别
    # Langfilters = ["ja_en"]           # 日英混合识别
    # Langfilters = ["zh_ko"]           # 中韩混合识别
    # Langfilters = ["ja_ko"]           # 日韩混合识别
    # Langfilters = ["en_ko"]           # 英韩混合识别
    # Langfilters = ["zh_ja_en"]        # 中日英混合识别
    # Langfilters = ["zh_ja_en_ko"]     # 中日英韩混合识别

    # 更多过滤组合，请您随意。。。For more filter combinations, please feel free to......


    # DEFINITION
    PARSE_TAG = re.compile(r'(⑥\$\d+[\d]{6,}⑥)')

    @staticmethod
    def _clears():
        LangSegment._text_cache = None
        LangSegment._text_lasts = None
        LangSegment._text_langs = None
        LangSegment._text_waits = None
        LangSegment._lang_count = None
        LangSegment._lang_eos   = None
        pass

    @staticmethod
    def _is_english_word(word):
        return bool(re.match(r'^[a-zA-Z]+$', word))

    @staticmethod
    def _is_chinese(word):
        for char in word:
            if '\u4e00' <= char <= '\u9fff':
                return True
        return False

    @staticmethod
    def _is_japanese_kana(word):
        pattern = re.compile(r'[\u3040-\u309F\u30A0-\u30FF]+')
        matches = pattern.findall(word)
        return len(matches) > 0

    @staticmethod
    def _insert_english_uppercase(word):
        modified_text = re.sub(r'(?<!\b)([A-Z])', r' \1', word)
        modified_text = modified_text.strip('-')
        return modified_text + " "

    @staticmethod
    def _saveData(words,language:str,text:str):
        # Language word statistics
        lang_count = LangSegment._lang_count
        if lang_count is None:lang_count = defaultdict(int)
        if not "|" in language:lang_count[language] += int(len(text)//2) if language == "en" else len(text)
        LangSegment._lang_count = lang_count
        # Merge the same language and save the results
        preData = words[-1] if len(words) > 0 else None
        if preData and  (preData["lang"] == language):
            text = preData["text"] + text
            preData["text"] = text
            return preData
        data = {"lang":language,"text": text}
        filters = LangSegment.Langfilters
        if filters is None or len(filters) == 0 or "?" in language or   \
            language in filters or language in filters[0] or \
            filters[0] == "*" or filters[0] in "alls-mixs-autos":
            words.append(data)
        return data

    @staticmethod
    def _addwords(words,language,text):
        if text is None or len(text.strip()) == 0:return True
        if language is None:language = ""
        language = language.lower()
        if language == 'en':text = LangSegment._insert_english_uppercase(text)
        # text = re.sub(r'[(（）)]', ',' , text) # Keep it.
        text_waits = LangSegment._text_waits
        ispre_waits = len(text_waits)>0
        preResult = text_waits.pop() if ispre_waits else None
        if preResult is None:preResult = words[-1] if len(words) > 0 else None
        if preResult and ("|" in preResult["lang"]):
            pre_lang = preResult["lang"]
            if language in pre_lang:preResult["lang"] = language = language.split("|")[0]
            else:preResult["lang"]=pre_lang.split("|")[0]
            if ispre_waits:preResult = LangSegment._saveData(words,preResult["lang"],preResult["text"])
        pre_lang = preResult["lang"] if preResult else None
        if ("|" in language) and (pre_lang and not pre_lang in language and not "…" in language):language = language.split("|")[0]
        filters = LangSegment.Langfilters
        if "|" in language:LangSegment._text_waits.append({"lang":language,"text": text})
        else:LangSegment._saveData(words,language,text)
        return False

    @staticmethod
    def _get_prev_data(words):
        data = words[-1] if words and len(words) > 0 else None
        if data:return (data["lang"] , data["text"])
        return (None,"")

    @staticmethod
    def _match_ending(input , index):
        if input is None or len(input) == 0:return False,None
        input = re.sub(r'\s+', '', input)
        if len(input) == 0 or abs(index) > len(input):return False,None
        ending_pattern = re.compile(r'([「」“”‘’"\':：。.！!?．？])')
        return ending_pattern.match(input[index]),input[index]

    @staticmethod
    def _cleans_text(cleans_text):
        cleans_text = re.sub(r'([^\w]+)', '', cleans_text)
        return cleans_text

    @staticmethod
    def _lang_classify(cleans_text):
        language, *_ = langid.classify(cleans_text)
        return language

    @staticmethod
    def _parse_language(words , segment):
        LANG_JA = "ja"
        LANG_ZH = "zh"
        language = LANG_ZH
        regex_pattern = re.compile(r'([^\w\s]+)')
        lines = regex_pattern.split(segment)
        lines_max = len(lines)
        LANG_EOS =LangSegment._lang_eos
        for index, text in enumerate(lines):
            if len(text) == 0:continue
            EOS = index >= (lines_max - 1)
            nextId = index + 1
            nextText = lines[nextId] if not EOS else ""
            nextPunc = len(re.sub(regex_pattern,'',re.sub(r'\n+','',nextText)).strip()) == 0
            textPunc = len(re.sub(regex_pattern,'',re.sub(r'\n+','',text)).strip()) == 0
            if not EOS and (textPunc == True or ( len(nextText.strip()) >= 0 and nextPunc == True)):
                lines[nextId] = f'{text}{nextText}'
                continue
            number_tags = re.compile(r'(⑥\d{6,}⑥)')
            cleans_text = re.sub(number_tags, '' ,text)
            cleans_text = LangSegment._cleans_text(cleans_text)
            language = LangSegment._lang_classify(cleans_text)
            prev_language , prev_text = LangSegment._get_prev_data(words)
            if len(cleans_text) <= 3 and LangSegment._is_chinese(cleans_text):
                if EOS and LANG_EOS: language = LANG_ZH if len(cleans_text) <= 1 else language
                elif LangSegment._is_japanese_kana(cleans_text):language = LANG_JA
                else:
                    LANG_UNKNOWN = f'{LANG_ZH}|{LANG_JA}'
                    match_end,match_char = LangSegment._match_ending(text, -1)
                    referen = prev_language in LANG_UNKNOWN or LANG_UNKNOWN in prev_language if prev_language else False
                    if match_char in "。.": language = prev_language if referen and len(words) > 0 else language
                    else:language = f"{LANG_UNKNOWN}|…"
            text,*_ = re.subn(number_tags , LangSegment._restore_number , text )
            LangSegment._addwords(words,language,text)
            pass
        pass

    @staticmethod
    def _restore_number(matche):
        value = matche.group(0)
        text_cache = LangSegment._text_cache
        if value in text_cache:
            process , data = text_cache[value]
            tag , match = data
            value = match
        return value

    @staticmethod
    def _pattern_symbols(item , text):
        if text is None:return text
        tag , pattern , process = item
        matches = pattern.findall(text)
        if len(matches) == 1 and "".join(matches[0]) == text:
            return text
        for i , match in enumerate(matches):
            key = f"⑥{tag}{i:06d}⑥"
            text = re.sub(pattern , key , text , count=1)
            LangSegment._text_cache[key] = (process , (tag , match))
        return text

    @staticmethod
    def _process_symbol(words,data):
        tag , match = data
        language = match[1]
        text = match[2]
        LangSegment._addwords(words,language,text)
        pass

    @staticmethod
    def _process_english(words,data):
        tag , match = data
        text = match[0]
        language = "en"
        LangSegment._addwords(words,language,text)
        pass

    @staticmethod
    def _process_korean(words,data):
        tag , match = data
        text = match[0]
        language = "ko"
        LangSegment._addwords(words,language,text)
        pass

    @staticmethod
    def _process_quotes(words,data):
        tag , match = data
        text = "".join(match)
        childs = LangSegment.PARSE_TAG.findall(text)
        if len(childs) > 0:
            LangSegment._process_tags(words , text , False)
        else:
            cleans_text = LangSegment._cleans_text(match[1])
            if len(cleans_text) <= 3:
                LangSegment._parse_language(words,text)
            else:
                language = LangSegment._lang_classify(cleans_text)
                LangSegment._addwords(words,language,text)
        pass

    @staticmethod
    def _process_number(words,data): # "$0" process only
        """
        Numbers alone cannot accurately identify language.
        Because numbers are universal in all languages.
        So it won't be executed here, just for testing.
        """
        tag , match = data
        language = words[0]["lang"] if len(words) > 0 else "zh"
        text = match
        LangSegment._addwords(words,language,text)
        pass

    @staticmethod
    def _process_tags(words , text , root_tag):
        text_cache = LangSegment._text_cache
        segments = re.split(LangSegment.PARSE_TAG, text)
        segments_len = len(segments) - 1
        for index , text in enumerate(segments):
            if root_tag:LangSegment._lang_eos = index >= segments_len
            if LangSegment.PARSE_TAG.match(text):
                process , data = text_cache[text]
                if process:process(words , data)
            else:
                LangSegment._parse_language(words , text)
            pass
        return words

    @staticmethod
    def _parse_symbols(text):
        TAG_NUM = "00" # "00" => default channels , "$0" => testing channel
        TAG_S1,TAG_P1,TAG_P2,TAG_EN,TAG_KO = "$1" ,"$2" ,"$3" ,"$4" ,"$5"
        process_list = [
            (  TAG_S1  , re.compile(LangSegment.SYMBOLS_PATTERN) , LangSegment._process_symbol  ),      # Symbol Tag
            (  TAG_KO  , re.compile('(([【《（(“‘"\']*(\d+\W*\s*)*[\uac00-\ud7a3]+[\W\s]*)+)')  , LangSegment._process_korean  ),      # Korean words
            (  TAG_NUM , re.compile(r'(\W*\d+\W+\d*\W*\d*)')        , LangSegment._process_number  ),      # Number words, Universal in all languages, Ignore it.
            (  TAG_EN  , re.compile(r'(([【《（(“‘"\']*[a-zA-Z]+[\W\s]*)+)')    , LangSegment._process_english ),                      # English words
            (  TAG_P1  , re.compile(r'(["\'])(.*?)(\1)')         , LangSegment._process_quotes  ),      # Regular quotes
            (  TAG_P2  , re.compile(r'([\n]*[【《（(“‘])([^【《（(“‘’”)）》】]{3,})([’”)）》】][\W\s]*[\n]{,1})')   , LangSegment._process_quotes  ),  # Special quotes, There are left and right.
        ]
        LangSegment._lang_eos = False
        text_cache = LangSegment._text_cache = {}
        for item in process_list:
            text = LangSegment._pattern_symbols(item , text)
        words = LangSegment._process_tags([] , text , True)
        lang_count = LangSegment._lang_count
        if lang_count and len(lang_count) > 0:
            lang_count = dict(sorted(lang_count.items(), key=lambda x: x[1], reverse=True))
            lang_count = list(lang_count.items())
            LangSegment._lang_count = lang_count
        return words

    @staticmethod
    def setfilters(filters):
        # 当过滤器更改时，清除缓存
        # When the filter changes, clear the cache
        if LangSegment.Langfilters != filters:
            LangSegment._clears()
            LangSegment.Langfilters = filters
        pass

    @staticmethod
    def getfilters():
        return LangSegment.Langfilters


    @staticmethod
    def getCounts():
        lang_count = LangSegment._lang_count
        if lang_count is not None:return lang_count
        text_langs = LangSegment._text_langs
        if text_langs is None or len(text_langs) == 0:return [("zh",0)]
        lang_counts = defaultdict(int)
        for d in text_langs:lang_counts[d['lang']] += int(len(d['text'])//2) if d['lang'] == "en" else len(d['text'])
        lang_counts = dict(sorted(lang_counts.items(), key=lambda x: x[1], reverse=True))
        lang_counts = list(lang_counts.items())
        LangSegment._lang_count = lang_counts
        return lang_counts

    @staticmethod
    def getTexts(text:str):
        if text is None or len(text.strip()) == 0:
            LangSegment._clears()
            return []
        # lasts
        text_langs = LangSegment._text_langs
        if LangSegment._text_lasts == text and text_langs is not None:return text_langs
        # parse
        LangSegment._text_waits = []
        LangSegment._lang_count = None
        LangSegment._text_lasts = text
        text = LangSegment._parse_symbols(text)
        LangSegment._text_langs = text
        return text

    @staticmethod
    def classify(text:str):
        return LangSegment.getTexts(text)

def setfilters(filters):
    """
    功能：语言过滤组功能, 可以指定保留语言。不在过滤组中的语言将被清除。您可随心搭配TTS语音合成所支持的语言。
    Function: Language filter group function, you can specify reserved languages. \n
    Languages not in the filter group will be cleared. You can match the languages supported by TTS Text To Speech as you like.\n
    Args:
        filters (list): ["zh", "en", "ja", "ko"]
    """
    LangSegment.setfilters(filters)
    pass

def getfilters():
    """
    功能：语言过滤组功能, 可以指定保留语言。不在过滤组中的语言将被清除。您可随心搭配TTS语音合成所支持的语言。
    Function: Language filter group function, you can specify reserved languages. \n
    Languages not in the filter group will be cleared. You can match the languages supported by TTS Text To Speech as you like.\n
    Args:
        filters (list): ["zh", "en", "ja", "ko"]
    """
    return LangSegment.getfilters()

# @Deprecated：Use shorter setfilters
def setLangfilters(filters):
    """
    >0.1.9废除：使用更简短的setfilters
    """
    setfilters(filters)
# @Deprecated：Use shorter getfilters
def getLangfilters():
    """
    >0.1.9废除：使用更简短的getfilters
    """
    return getfilters()

def getTexts(text:str):
    """
    功能：对输入的文本进行多语种分词\n
    Feature: Tokenizing multilingual text input.\n
    参数-Args:
        text (str): Text content,文本内容\n
    返回-Returns:
        list: 示例结果：[{'lang':'zh','text':'?'},...]\n
        lang=语种 , text=内容\n
    """
    return LangSegment.getTexts(text)

def getCounts():
    """
    功能：分词结果统计，按语种字数降序，用于确定其主要语言\n
    Function: Tokenizing multilingual text input.\n
    返回-Returns:
        list: 示例结果：[('zh', 5), ('ja', 2), ('en', 1)] = [(语种,字数含标点)]\n
    """
    return LangSegment.getCounts()

def classify(text:str):
    """
    功能：兼容接口实现
    Function: Compatible interface implementation
    """
    return LangSegment.classify(text)

def printList(langlist):
    """
    功能：打印数组结果
    Function: Print array results
    """
    print("\n\n===================【打印结果】===================")
    if langlist is None or len(langlist) == 0:
        print("无内容结果,No content result")
        return
    for line in langlist:
        print(line)
    pass



if __name__ == "__main__":

    # -----------------------------------
    # 更新日志：新版本分词更加精准。
    # Changelog: The new version of the word segmentation is more accurate.
    # -----------------------------------

    # 输入示例1：（包含日文，中文）
    # text = "“昨日は雨が降った，音楽、映画。。。”你今天学习日语了吗？春は桜の季節です。语种分词是语音合成必不可少的环节。言語分詞は音声合成に欠かせない環節である！"

    # 输入示例2：（包含日文，中文）
    # text = "欢迎来玩。東京，は日本の首都です。欢迎来玩.  太好了!"

    # 输入示例3：（包含日文，中文）
    # text = "明日、私たちは海辺にバカンスに行きます。你会说日语吗：“中国語、話せますか” 你的日语真好啊！"

    # 输入示例4：（包含日文，中文，韩语，英文）
    text = "你的名字叫<ja>佐々木？<ja>吗？韩语中的안녕 오빠读什么呢？あなたの体育の先生は誰ですか? 此次发布会带来了四款iPhone 15系列机型和三款Apple Watch等一系列新品，这次的iPad Air采用了LCD屏幕"

    # 进行分词：（接入TTS项目仅需一行代码调用）
    langlist = LangSegment.getTexts(text)
    printList(langlist)


    # 语种统计:
    print("\n===================【语种统计】===================")
    # 获取所有语种数组结果，根据内容字数降序排列
    langCounts = LangSegment.getCounts()
    print(langCounts , "\n")

    # 根据结果获取内容的主要语种 (语言，字数含标点)
    lang , count = langCounts[0]
    print(f"输入内容的主要语言为 = {lang} ，字数 = {count}")
    print("==================================================\n")


    # 分词输出：lang=语言，text=内容
    # ===================【打印结果】===================
    # {'lang': 'zh', 'text': '你的名字叫'}
    # {'lang': 'ja', 'text': '佐々木？'}
    # {'lang': 'zh', 'text': '吗？韩语中的'}
    # {'lang': 'ko', 'text': '안녕 오빠'}
    # {'lang': 'zh', 'text': '读什么呢？'}
    # {'lang': 'ja', 'text': 'あなたの体育の先生は誰ですか?'}
    # {'lang': 'zh', 'text': ' 此次发布会带来了四款'}
    # {'lang': 'en', 'text': 'i Phone  '}
    # {'lang': 'zh', 'text': '15系列机型和三款'}
    # {'lang': 'en', 'text': 'Apple Watch '}
    # {'lang': 'zh', 'text': '等一系列新品，这次的'}
    # {'lang': 'en', 'text': 'i Pad Air '}
    # {'lang': 'zh', 'text': '采用了'}
    # {'lang': 'en', 'text': 'L C D '}
    # {'lang': 'zh', 'text': '屏幕'}
    # ===================【语种统计】===================

    # ===================【语种统计】===================
    # [('zh', 51), ('ja', 19), ('en', 18), ('ko', 5)]

    # 输入内容的主要语言为 = zh ，字数 = 51
    # ==================================================






import subprocess


def stop_process(process):
    """Terminate one or many subprocess.Popen objects gracefully."""
    processes = process if isinstance(process, list) else [process]
    for proc in processes:
        if proc is None:
            continue
        try:
            if proc.poll() is None:
                proc.terminate()
                try:
                    proc.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    proc.kill()
        except Exception:
            # Best-effort shutdown; ignore failures so cleanup keeps going
            continue

def stop_dora_dataflow(dataflow_name:str):
    dora_stop_process = subprocess.Popen(
        ['dora', 'stop','--name',dataflow_name],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    stdout, stderr = dora_stop_process.communicate()
    return None


def destroy_dora_daemon():
    """Best-effort wrapper around `dora destroy` to cleanup daemon state."""
    try:
        subprocess.run(
            ['dora', 'destroy'],
            check=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except Exception:
        # If destroy fails we silently ignore so CLI shutdown still succeeds
        pass

def send_task_or_stop_process(task:str,dora_dataflow_process,task_input_process,dataflow_name:str):
    if task.lower() in ["exit", "quit"]:
        stop_process([dora_dataflow_process, task_input_process])
        stop_dora_dataflow(dataflow_name=dataflow_name)
        return False
    if task_input_process.poll() is None:
        task_input_process.stdin.write(task + '\n')
        task_input_process.stdin.flush()
        return True








import re
from typing import Callable

punctuation = set(['!', '?', '…', ',', '.', '-'," "])
METHODS = dict()

def get_method(name:str)->Callable:
    method = METHODS.get(name, None)
    if method is None:
        raise ValueError(f"Method {name} not found")
    return method

def get_method_names()->list:
    return list(METHODS.keys())

def register_method(name):
    def decorator(func):
        METHODS[name] = func
        return func
    return decorator

splits = {"，", "。", "？", "！", ",", ".", "?", "!", "~", ":", "：", "—", "…", }

def split_big_text(text, max_len=510):
    # 定义全角和半角标点符号
    punctuation = "".join(splits)

    # 切割文本
    segments = re.split('([' + punctuation + '])', text)
    
    # 初始化结果列表和当前片段
    result = []
    current_segment = ''
    
    for segment in segments:
        # 如果当前片段加上新的片段长度超过max_len，就将当前片段加入结果列表，并重置当前片段
        if len(current_segment + segment) > max_len:
            result.append(current_segment)
            current_segment = segment
        else:
            current_segment += segment
    
    # 将最后一个片段加入结果列表
    if current_segment:
        result.append(current_segment)
    
    return result



def split(todo_text):
    todo_text = todo_text.replace("……", "。").replace("——", "，")
    if todo_text[-1] not in splits:
        todo_text += "。"
    i_split_head = i_split_tail = 0
    len_text = len(todo_text)
    todo_texts = []
    while 1:
        if i_split_head >= len_text:
            break  # 结尾一定有标点，所以直接跳出即可，最后一段在上次已加入
        if todo_text[i_split_head] in splits:
            i_split_head += 1
            todo_texts.append(todo_text[i_split_tail:i_split_head])
            i_split_tail = i_split_head
        else:
            i_split_head += 1
    return todo_texts


# 不切
@register_method("cut0")
def cut0(inp):
    if not set(inp).issubset(punctuation):
        return inp
    else:
        return "/n"


# 凑四句一切
@register_method("cut1")
def cut1(inp):
    inp = inp.strip("\n")
    inps = split(inp)
    split_idx = list(range(0, len(inps), 4))
    split_idx[-1] = None
    if len(split_idx) > 1:
        opts = []
        for idx in range(len(split_idx) - 1):
            opts.append("".join(inps[split_idx[idx]: split_idx[idx + 1]]))
    else:
        opts = [inp]
    opts = [item for item in opts if not set(item).issubset(punctuation)]
    return "\n".join(opts)


# 凑50字一切
@register_method("cut2")
def cut2(inp):
    inp = inp.strip("\n")
    inps = split(inp)
    if len(inps) < 2:
        return inp
    opts = []
    summ = 0
    tmp_str = ""
    for i in range(len(inps)):
        summ += len(inps[i])
        tmp_str += inps[i]
        if summ > 50:
            summ = 0
            opts.append(tmp_str)
            tmp_str = ""
    if tmp_str != "":
        opts.append(tmp_str)
    # print(opts)
    if len(opts) > 1 and len(opts[-1]) < 50:  ##如果最后一个太短了，和前一个合一起
        opts[-2] = opts[-2] + opts[-1]
        opts = opts[:-1]
    opts = [item for item in opts if not set(item).issubset(punctuation)]
    return "\n".join(opts)

# 按中文句号。切
@register_method("cut3")
def cut3(inp):
    inp = inp.strip("\n")
    opts = ["%s" % item for item in inp.strip("。").split("。")]
    opts = [item for item in opts if not set(item).issubset(punctuation)]
    return "\n".join(opts)

#按英文句号.切
@register_method("cut4")
def cut4(inp):
    inp = inp.strip("\n")
    opts = ["%s" % item for item in inp.strip(".").split(".")]
    opts = [item for item in opts if not set(item).issubset(punctuation)]
    return "\n".join(opts)

# 按标点符号切
# contributed by https://github.com/AI-Hobbyist/GPT-SoVITS/blob/main/GPT_SoVITS/inference_webui.py
@register_method("cut5")
def cut5(inp):
    inp = inp.strip("\n")
    punds = {',', '.', ';', '?', '!', '、', '，', '。', '？', '！', ';', '：', '…'}
    mergeitems = []
    items = []

    for i, char in enumerate(inp):
        if char in punds:
            if char == '.' and i > 0 and i < len(inp) - 1 and inp[i - 1].isdigit() and inp[i + 1].isdigit():
                items.append(char)
            else:
                items.append(char)
                mergeitems.append("".join(items))
                items = []
        else:
            items.append(char)

    if items:
        mergeitems.append("".join(items))

    opt = [item for item in mergeitems if not set(item).issubset(punds)]
    return "\n".join(opt)



if __name__ == '__main__':
    method = get_method("cut5")
    print(method("你好，我是小明。你好，我是小红。你好，我是小刚。你好，我是小张。"))
    



import ast
import os
import sys
import importlib
from typing import Dict, List, Set, Any


def remove_main_guard(raw_code):
    """Fixed code by removing the if __name__ == "__main__" guard and its main() call."""
    lines = raw_code.split('\n')
    new_lines = []
    skip = False
    # Support both spaced and unspaced versions
    target_patterns = [
        'if __name__=="__main__":',
        'if __name__ == "__main__":'
    ]
    
    for line in lines:
        stripped_line = line.strip()
        # Check for the main guard line
        if stripped_line in target_patterns:
            skip = True
            continue
        # If we are in a skip state, check for the main() call
        if skip:
            # Skip the main() call line and end the skip state afterwards
            if stripped_line == 'main()':
                skip = False
                continue
            # Skip empty lines or other indented lines (adjust based on actual code formatting)
            if not stripped_line:
                continue
        # If not skipping, keep the line
        new_lines.append(line)
    
    return '\n'.join(new_lines)

class DependencyAnalyzer(ast.NodeVisitor):
    def __init__(self):
        self.imported_modules = set()
        self.imported_objects = dict()
        self.global_functions = set()
        self.global_variables = set()
        self.relative_imports = set()

    def visit_Import(self, node):
        for alias in node.names:
            self.imported_modules.add(alias.name)
        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        module = node.module or ''
        for alias in node.names:
            self.imported_objects[alias.asname or alias.name] = (module, alias.name)
        if module:
            self.imported_modules.add(module)
        # Track relative imports
        if node.level > 0:
            self.relative_imports.add((node.level, module))
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self.global_functions.add(node.name)
        self.generic_visit(node)

    def visit_Assign(self, node):
        if isinstance(getattr(node, 'parent', None), ast.Module):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    self.global_variables.add(target.id)
        self.generic_visit(node)

class ParentSetter(ast.NodeVisitor):
    def visit(self, node):
        for child in ast.iter_child_nodes(node):
            child.parent = node
            self.visit(child)

def analyze_dependencies(code: str) -> Dict:
    tree = ast.parse(code)
    ParentSetter().visit(tree)
    analyzer = DependencyAnalyzer()
    analyzer.visit(tree)
    return {
        "imported_modules": analyzer.imported_modules,
        "imported_objects": analyzer.imported_objects,
        "global_functions": analyzer.global_functions,
        "global_variables": analyzer.global_variables,
        "relative_imports": analyzer.relative_imports
    }

def normalize_relative_imports(source_code: str) -> str:
    """
    Normalize relative imports (from .xxx / from ..xxx) by setting level to 0
    while keeping module name unchanged. This makes code exec-able when there is
    no package context (module_path is None). Falls back to original source on
    any failure.
    """
    try:
        tree = ast.parse(source_code)
        class RelImportRewriter(ast.NodeTransformer):
            def visit_ImportFrom(self, node):
                if getattr(node, 'level', 0) and node.module is not None:
                    # Drop relative level but keep module name
                    return ast.copy_location(
                        ast.ImportFrom(module=node.module, names=node.names, level=0),
                        node
                    )
                return node
        new_tree = RelImportRewriter().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)
    except Exception:
        return source_code

def build_execution_globals(source_code: str, module_path: str = None) -> Dict[str, Any]:
    execution_globals = {
        "__builtins__": __builtins__,
        "__name__": "__main__",
        "__file__": module_path or "",
        "__package__": os.path.basename(os.path.dirname(module_path)) if module_path else None  # set package name if path provided
    }
    deps = analyze_dependencies(source_code)
    
    # Handle relative imports when module_path is known
    if deps["relative_imports"] and module_path:
        module_dir = os.path.dirname(os.path.abspath(module_path))
        parent_dir = os.path.dirname(module_dir)
        if parent_dir not in sys.path:
            sys.path.insert(0, parent_dir)

    # Import modules
    for module_name in deps["imported_modules"]:
        module = None
        # 1) Try absolute import first
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            pass

        # 2) If failed and package context exists, try package-relative import
        if module is None and module_path and execution_globals.get("__package__"):
            try:
                module = importlib.import_module(module_name, package=execution_globals["__package__"])
            except ImportError:
                pass

        # 3) Finally try explicit dotted-relative form (from .xxx)
        if module is None and module_path and execution_globals.get("__package__"):
            try:
                module = importlib.import_module(f".{module_name}", package=execution_globals["__package__"])
            except ImportError:
                pass

        if module is not None:
            execution_globals[module_name] = module
        else:
            print(f"警告：模块 {module_name} 导入失败，可能导致执行错误")

    for alias, (module_name, obj_name) in deps["imported_objects"].items():
        module = None
        # If module was already loaded and contains the object, use it
        if module_name in execution_globals and hasattr(execution_globals[module_name], obj_name):
            execution_globals[alias] = getattr(execution_globals[module_name], obj_name)
            continue

        # 1) Absolute import
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            module = None

        # 2) Package-relative import
        if module is None and module_path and execution_globals.get("__package__"):
            try:
                module = importlib.import_module(module_name, package=execution_globals["__package__"])
            except ImportError:
                module = None

        # 3) Dotted-relative import
        if module is None and module_path and execution_globals.get("__package__"):
            try:
                module = importlib.import_module(f".{module_name}", package=execution_globals["__package__"])
            except ImportError:
                module = None

        if module is not None:
            try:
                execution_globals[alias] = getattr(module, obj_name)
            except AttributeError:
                print(f"警告：对象 {obj_name} 从模块 {module_name} 导入失败")
        else:
            print(f"警告：对象 {obj_name} 从模块 {module_name} 导入失败")

    # Load global functions and variables
    temp_globals = execution_globals.copy()
    try:
        code_to_exec = source_code
        # If module_path is not provided, normalize relative imports to avoid
        # "no known parent package"
        if not module_path and deps["relative_imports"]:
            code_to_exec = normalize_relative_imports(source_code)
        exec(code_to_exec, temp_globals)
        for func_name in deps["global_functions"]:
            if func_name in temp_globals:
                execution_globals[func_name] = temp_globals[func_name]
        for var_name in deps["global_variables"]:
            if var_name in temp_globals:
                execution_globals[var_name] = temp_globals[var_name]
    except Exception as e:
        print(f"Error while executing source code to load globals: {e}")

    return execution_globals

def extract_agent_info(code: str) -> Dict:
    tree = ast.parse(code)
    
    receive_nodes = []
    send_nodes = []
    
    class AgentCallVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            if isinstance(node.value, ast.Call):
                self._check_call_node(node.value, node)
            self.generic_visit(node)
            
        def visit_Expr(self, node):
            if isinstance(node.value, ast.Call):
                self._check_call_node(node.value, node)
            self.generic_visit(node)
            
        def _check_call_node(self, call_node: ast.Call, parent_node):
            if (isinstance(call_node.func, ast.Attribute) and
                isinstance(call_node.func.value, ast.Name) and
                call_node.func.value.id == 'agent'):
                
                func_name = call_node.func.attr
                # Accept common variants: singular, plural and short names
                if func_name in ('receive_parameter', 'receive_parameters', 'receive_params', 'receive'):
                    receive_nodes.append((parent_node, call_node))
                elif func_name == 'send_output':
                    send_nodes.append((parent_node, call_node))
    
    visitor = AgentCallVisitor()
    visitor.visit(tree)
    
    def get_call_args(call_node: ast.Call) -> List[str]:
        args = []
        for arg in call_node.args:
            # If the argument is a list/tuple literal of string constants, expand it
            if isinstance(arg, (ast.List, ast.Tuple)):
                all_consts = True
                for elt in arg.elts:
                    if not isinstance(elt, ast.Constant) or not isinstance(elt.value, str):
                        all_consts = False
                        break
                if all_consts:
                    for elt in arg.elts:
                        args.append(f"'{elt.value}'")
                    continue

            if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                args.append(f"'{arg.value}'")
            else:
                args.append(ast.unparse(arg))
        for kw in call_node.keywords:
            value_str = ast.unparse(kw.value)
            if isinstance(kw.value, ast.Constant) and isinstance(kw.value.value, str):
                value_str = f"'{kw.value.value}'"
            args.append(f"{kw.arg}={value_str}")
        return args
    
    receive_params = []
    if receive_nodes:
        receive_params = get_call_args(receive_nodes[0][1])
    
    receive_target = None
    if receive_nodes:
        parent = receive_nodes[0][0]
        try:
            if isinstance(parent, ast.Assign):
                targets = parent.targets
                if targets:
                    t = targets[0]
                    if isinstance(t, ast.Name):
                        receive_target = t.id
                    elif isinstance(t, ast.Tuple):
                        for elt in t.elts:
                            if isinstance(elt, ast.Name):
                                receive_target = elt.id
                                break
            elif isinstance(parent, ast.AnnAssign):
                t = parent.target
                if isinstance(t, ast.Name):
                    receive_target = t.id
        except Exception:
            receive_target = None
    
    send_params = []
    if send_nodes:
        send_params = get_call_args(send_nodes[0][1])
    
    between_code = ""
    if receive_nodes and send_nodes:
        receive_parent, _ = receive_nodes[0]
        send_parent, _ = send_nodes[0]
        
        if receive_parent.lineno < send_parent.lineno:
            start_line = receive_parent.end_lineno + 1
            end_line = send_parent.lineno - 1
            
            if start_line <= end_line:
                code_lines = code.split('\n')
                between_lines = []
                for line_num in range(start_line-1, end_line):
                    if line_num < len(code_lines):
                        line = code_lines[line_num].rstrip('\r')
                        if line.strip():
                            between_lines.append(line)
                
                if between_lines:
                    min_indent = min(len(line) - len(line.lstrip()) for line in between_lines)
                    between_lines = [line[min_indent:] for line in between_lines]
                    between_code = '\n'.join(between_lines).strip()
    
    return {
        "receive_params": receive_params,
        "receive_target": receive_target,
        "send_params": send_params,
        "between_code": between_code
    }

def load_node_module(node_folder_path: str, execute: bool = False):
    node_folder_path = os.path.abspath(node_folder_path)

    if not os.path.exists(node_folder_path):
        raise ImportError(f"Node folder not found: {node_folder_path}")

    candidate = None
    for root, dirs, files in os.walk(node_folder_path):
        if 'main.py' in files:
            candidate = os.path.join(root, 'main.py')
            break
        for f in files:
            if f.endswith('.py'):
                candidate = os.path.join(root, f)

    if candidate is None:
        try:
            for f in os.listdir(node_folder_path):
                if f.endswith('.py'):
                    candidate = os.path.join(node_folder_path, f)
                    break
        except Exception:
            pass

    if candidate is None:
        raise ImportError(f"No python module found in {node_folder_path}")

    try:
        with open(candidate, 'r', encoding='utf-8') as _f:
            source_code = _f.read()
    except Exception as e:
        raise ImportError(f"Failed to read source file {candidate}: {e}") from e

    agent_info = None
    try:
        parsed = ast.parse(source_code)
        has_main = any(isinstance(n, ast.FunctionDef) and n.name == 'main' for n in parsed.body)
        if has_main:
            try:
                agent_info = extract_agent_info(source_code)
            except Exception:
                agent_info = None
    except Exception:
        agent_info = None

    execution_globals = None
    if execute:
        execution_globals = build_execution_globals(source_code, candidate)

    descriptor = {
        'path': candidate,
        'source': source_code,
        'agent_info': agent_info,
        'execution_globals': execution_globals
    }

    if not execute:
        return descriptor
    
    try:
        exec(source_code, execution_globals)
    except Exception as e:
        print(f"Error executing module: {e}")
    return descriptor

if __name__ == "__main__":
    sample_code = """
from mofa.agent_build.base.base_agent import MofaAgent, run_agent
from openai import OpenAI
import os

from .multi_llm import choice_and_run_llm_model  # 相对导入

def identify_info_types(user_input):
    info_types = []
    medicine_keywords = ["药", "吃药"]
    if any(keyword in user_input for keyword in medicine_keywords):
        info_types.append("用药")
    return info_types if info_types else ["其他"]

@run_agent
def run(agent: MofaAgent):
    user_query = agent.receive_parameter('query')
    types = identify_info_types(user_query)
    result = f"识别到信息类型: {types}"
    agent.send_output(agent_output_name='result', agent_result=result)
    """
    
    deps = analyze_dependencies(sample_code)
    print("Imported modules:", deps["imported_modules"])
    print("Imported objects:", deps["imported_objects"])
    print("Global functions:", deps["global_functions"])
    
    exec_globals = build_execution_globals(sample_code)
    print("\nExecution globals contain:", list(exec_globals.keys()))
    
    agent_info = extract_agent_info(sample_code)
    print("\nBetween code:\n", agent_info["between_code"])
    
    if agent_info["between_code"]:
        try:
            exec_globals["agent"] = None
            exec_globals["user_query"] = "我需要吃什么药？"
            
            # Fix typo: use exec_globals
            if "identify_info_types" not in exec_globals:
                print("Warning: identify_info_types not loaded, trying reload...")
                temp_globals = {**exec_globals}
                # Normalize relative imports for demo code
                demo_code = normalize_relative_imports(sample_code)
                exec(demo_code, temp_globals)
                exec_globals["identify_info_types"] = temp_globals.get("identify_info_types")
            
            exec(agent_info["between_code"], exec_globals)
            print("Execute success, result:", exec_globals.get("result"))
        except Exception as e:
            print("Error executing between code:", e)

__all__ = ["extract_agent_info", "load_node_module", "analyze_dependencies", "build_execution_globals"]



# This code is modified from https://github.com/PaddlePaddle/PaddleSpeech/tree/develop/paddlespeech/t2s/frontend/g2pw
# This code is modified from https://github.com/GitYCC/g2pW

import warnings
warnings.filterwarnings("ignore")
import json
import os
import zipfile,requests
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import numpy as np
import onnxruntime
onnxruntime.set_default_logger_severity(3)
from opencc import OpenCC
from transformers import AutoTokenizer
from pypinyin import pinyin
from pypinyin import Style

from .dataset import get_char_phoneme_labels
from .dataset import get_phoneme_labels
from .dataset import prepare_onnx_input
from .utils import load_config
from ..zh_normalization.char_convert import tranditional_to_simplified

model_version = '1.1'


def predict(session, onnx_input: Dict[str, Any],
            labels: List[str]) -> Tuple[List[str], List[float]]:
    all_preds = []
    all_confidences = []
    probs = session.run([], {
        "input_ids": onnx_input['input_ids'],
        "token_type_ids": onnx_input['token_type_ids'],
        "attention_mask": onnx_input['attention_masks'],
        "phoneme_mask": onnx_input['phoneme_masks'],
        "char_ids": onnx_input['char_ids'],
        "position_ids": onnx_input['position_ids']
    })[0]

    preds = np.argmax(probs, axis=1).tolist()
    max_probs = []
    for index, arr in zip(preds, probs.tolist()):
        max_probs.append(arr[index])
    all_preds += [labels[pred] for pred in preds]
    all_confidences += max_probs

    return all_preds, all_confidences


def download_and_decompress(model_dir: str='G2PWModel/'):
    if not os.path.exists(model_dir):
        parent_directory = os.path.dirname(model_dir)
        zip_dir = os.path.join(parent_directory,"G2PWModel_1.1.zip")
        extract_dir = os.path.join(parent_directory,"G2PWModel_1.1")
        extract_dir_new = os.path.join(parent_directory,"G2PWModel")
        print("Downloading g2pw model...")
        modelscope_url = "https://paddlespeech.bj.bcebos.com/Parakeet/released_models/g2p/G2PWModel_1.1.zip"
        with requests.get(modelscope_url, stream=True) as r:
            r.raise_for_status()
            with open(zip_dir, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)

        print("Extracting g2pw model...")
        with zipfile.ZipFile(zip_dir, "r") as zip_ref:
            zip_ref.extractall(parent_directory)
        
        os.rename(extract_dir, extract_dir_new)

    return model_dir

class G2PWOnnxConverter:
    def __init__(self,
                 model_dir: str='G2PWModel/',
                 style: str='bopomofo',
                 model_source: str=None,
                 enable_non_tradional_chinese: bool=False):
        uncompress_path = download_and_decompress(model_dir)

        sess_options = onnxruntime.SessionOptions()
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        sess_options.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
        sess_options.intra_op_num_threads = 2
        try:
            self.session_g2pW = onnxruntime.InferenceSession(os.path.join(uncompress_path, 'g2pW.onnx'),sess_options=sess_options, providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])
        except:
            self.session_g2pW = onnxruntime.InferenceSession(os.path.join(uncompress_path, 'g2pW.onnx'),sess_options=sess_options, providers=['CPUExecutionProvider'])
        self.config = load_config(
            config_path=os.path.join(uncompress_path, 'config.py'),
            use_default=True)

        self.model_source = model_source if model_source else self.config.model_source
        self.enable_opencc = enable_non_tradional_chinese

        self.tokenizer = AutoTokenizer.from_pretrained(self.model_source)

        polyphonic_chars_path = os.path.join(uncompress_path,
                                             'POLYPHONIC_CHARS.txt')
        monophonic_chars_path = os.path.join(uncompress_path,
                                             'MONOPHONIC_CHARS.txt')
        self.polyphonic_chars = [
            line.split('\t')
            for line in open(polyphonic_chars_path, encoding='utf-8').read()
            .strip().split('\n')
        ]
        self.non_polyphonic = {
            '一', '不', '和', '咋', '嗲', '剖', '差', '攢', '倒', '難', '奔', '勁', '拗',
            '肖', '瘙', '誒', '泊', '听', '噢'
        }
        self.non_monophonic = {'似', '攢'}
        self.monophonic_chars = [
            line.split('\t')
            for line in open(monophonic_chars_path, encoding='utf-8').read()
            .strip().split('\n')
        ]
        self.labels, self.char2phonemes = get_char_phoneme_labels(
            polyphonic_chars=self.polyphonic_chars
        ) if self.config.use_char_phoneme else get_phoneme_labels(
            polyphonic_chars=self.polyphonic_chars)

        self.chars = sorted(list(self.char2phonemes.keys()))

        self.polyphonic_chars_new = set(self.chars)
        for char in self.non_polyphonic:
            if char in self.polyphonic_chars_new:
                self.polyphonic_chars_new.remove(char)

        self.monophonic_chars_dict = {
            char: phoneme
            for char, phoneme in self.monophonic_chars
        }
        for char in self.non_monophonic:
            if char in self.monophonic_chars_dict:
                self.monophonic_chars_dict.pop(char)

        self.pos_tags = [
            'UNK', 'A', 'C', 'D', 'I', 'N', 'P', 'T', 'V', 'DE', 'SHI'
        ]

        with open(
                os.path.join(uncompress_path,
                             'bopomofo_to_pinyin_wo_tune_dict.json'),
                'r',
                encoding='utf-8') as fr:
            self.bopomofo_convert_dict = json.load(fr)
        self.style_convert_func = {
            'bopomofo': lambda x: x,
            'pinyin': self._convert_bopomofo_to_pinyin,
        }[style]

        with open(
                os.path.join(uncompress_path, 'char_bopomofo_dict.json'),
                'r',
                encoding='utf-8') as fr:
            self.char_bopomofo_dict = json.load(fr)

        if self.enable_opencc:
            self.cc = OpenCC('s2tw')

    def _convert_bopomofo_to_pinyin(self, bopomofo: str) -> str:
        tone = bopomofo[-1]
        assert tone in '12345'
        component = self.bopomofo_convert_dict.get(bopomofo[:-1])
        if component:
            return component + tone
        else:
            print(f'Warning: "{bopomofo}" cannot convert to pinyin')
            return None

    def __call__(self, sentences: List[str]) -> List[List[str]]:
        if isinstance(sentences, str):
            sentences = [sentences]

        if self.enable_opencc:
            translated_sentences = []
            for sent in sentences:
                translated_sent = self.cc.convert(sent)
                assert len(translated_sent) == len(sent)
                translated_sentences.append(translated_sent)
            sentences = translated_sentences

        texts, query_ids, sent_ids, partial_results = self._prepare_data(
            sentences=sentences)
        if len(texts) == 0:
            # sentences no polyphonic words
            return partial_results

        onnx_input = prepare_onnx_input(
            tokenizer=self.tokenizer,
            labels=self.labels,
            char2phonemes=self.char2phonemes,
            chars=self.chars,
            texts=texts,
            query_ids=query_ids,
            use_mask=self.config.use_mask,
            window_size=None)

        preds, confidences = predict(
            session=self.session_g2pW,
            onnx_input=onnx_input,
            labels=self.labels)
        if self.config.use_char_phoneme:
            preds = [pred.split(' ')[1] for pred in preds]

        results = partial_results
        for sent_id, query_id, pred in zip(sent_ids, query_ids, preds):
            results[sent_id][query_id] = self.style_convert_func(pred)

        return results

    def _prepare_data(
            self, sentences: List[str]
    ) -> Tuple[List[str], List[int], List[int], List[List[str]]]:
        texts, query_ids, sent_ids, partial_results = [], [], [], []
        for sent_id, sent in enumerate(sentences):
            # pypinyin works well for Simplified Chinese than Traditional Chinese
            sent_s = tranditional_to_simplified(sent)
            pypinyin_result = pinyin(
                sent_s, neutral_tone_with_five=True, style=Style.TONE3)
            partial_result = [None] * len(sent)
            for i, char in enumerate(sent):
                if char in self.polyphonic_chars_new:
                    texts.append(sent)
                    query_ids.append(i)
                    sent_ids.append(sent_id)
                elif char in self.monophonic_chars_dict:
                    partial_result[i] = self.style_convert_func(
                        self.monophonic_chars_dict[char])
                elif char in self.char_bopomofo_dict:
                    partial_result[i] = pypinyin_result[i][0]
                    # partial_result[i] =  self.style_convert_func(self.char_bopomofo_dict[char][0])
                else:
                    partial_result[i] = pypinyin_result[i][0]

            partial_results.append(partial_result)
        return texts, query_ids, sent_ids, partial_results



import yaml

def parse_test_cases(yaml_file_path):
    """
    Parse test case YAML file
    
    Parameters:
        yaml_file_path: Path to the YAML file
        
    Returns:
        Parsed list of test cases
    """
    try:
        with open(yaml_file_path, 'r', encoding='utf-8') as f:
            # Load YAML file content
            data = yaml.safe_load(f)
            
            # Check if test_cases field exists
            if 'test_cases' not in data:
                raise ValueError("test_cases field not found in YAML file")
                
            return data['test_cases']
            
    except FileNotFoundError:
        print(f"Error: File {yaml_file_path} not found")
        return None
    except yaml.YAMLError as e:
        print(f"YAML parsing error: {e}")
        return None

def main():
    # Assume the YAML file is named test_hello_world.yml
    test_cases = parse_test_cases('./tests/test_hello_world.yml')
    
    if test_cases:
        print(f"Successfully parsed {len(test_cases)} test cases:")
        for i, case in enumerate(test_cases, 1):
            print(f"\nTest case {i}: {case['name']}")
            print(f"Input: {case['input']}")
            print(f"Expected output: {case['expected_output']}")

if __name__ == "__main__":
    main()
    


# modified from https://github.com/lifeiteng/vall-e/blob/main/valle/modules/embedding.py
import math

import torch
from torch import nn


class TokenEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        vocab_size: int,
        dropout: float = 0.0,
    ):
        super().__init__()

        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim

        self.dropout = torch.nn.Dropout(p=dropout)
        self.word_embeddings = nn.Embedding(self.vocab_size, self.embedding_dim)

    @property
    def weight(self) -> torch.Tensor:
        return self.word_embeddings.weight

    def embedding(self, index: int) -> torch.Tensor:
        return self.word_embeddings.weight[index : index + 1]

    def forward(self, x: torch.Tensor):
        x = self.word_embeddings(x)
        x = self.dropout(x)
        return x


class SinePositionalEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        dropout: float = 0.0,
        scale: bool = False,
        alpha: bool = False,
    ):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.x_scale = math.sqrt(embedding_dim) if scale else 1.0
        self.alpha = nn.Parameter(torch.ones(1), requires_grad=alpha)
        self.dropout = torch.nn.Dropout(p=dropout)
        self.reverse = False
        self.div_term = torch.exp(torch.arange(0, self.embedding_dim, 2) * -(math.log(10000.0) / self.embedding_dim))

    def extend_pe(self, x):
        position = torch.cumsum(torch.ones_like(x[:,:,0]), dim=1).transpose(0, 1)
        scpe = (position * self.div_term).unsqueeze(0)
        pe = torch.cat([torch.sin(scpe), torch.cos(scpe)]).permute(1, 2, 0)
        pe = pe.contiguous().view(1, -1, self.embedding_dim)
        return pe

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pe = self.extend_pe(x)
        output = x.unsqueeze(-1) if x.ndim == 2 else x
        output = output * self.x_scale + self.alpha * pe
        return self.dropout(output)



#!/usr/bin/env python3
"""
Script Segmenter - Parse markdown and orchestrate TTS generation.
Streams text segments sequentially to downstream TTS nodes and
includes intelligent text segmentation for long passages.
"""
import argparse
import json
import time
import os
import re
from typing import Dict, List, Optional, Tuple, Iterable
import pyarrow as pa
from mofa.agent_build.base.base_agent import MofaAgent, run_agent


def send_log(node, level, message, config_level="INFO"):
    """Send log message through log output channel."""
    LOG_LEVELS = {"DEBUG": 10, "INFO": 20, "WARNING": 30, "ERROR": 40}

    if LOG_LEVELS.get(level, 0) < LOG_LEVELS.get(config_level, 20):
        return

    formatted_message = f"[{level}] {message}"
    log_data = {
        "node": "script-segmenter",
        "level": level,
        "message": formatted_message,
        "timestamp": time.time()
    }
    node.node.send_output("log", pa.array([json.dumps(log_data)]))


def find_split_index(text: str, max_length: int, split_marks: Iterable[str]) -> int:
    """Find a split index at or before max_length using provided marks or whitespace."""
    if max_length <= 0:
        return -1

    limit = min(len(text), max_length)

    # Try to split at punctuation marks first
    if split_marks:
        for idx in range(limit, 0, -1):
            if text[idx - 1] in split_marks:
                return idx

    # Fall back to whitespace
    for idx in range(limit, 0, -1):
        if text[idx - 1].isspace():
            return idx

    return -1


def split_long_text(
    text: str,
    max_length: int,
    punctuation_marks: str,
    node=None,
    log_level: str = "INFO",
) -> List[str]:
    """Split long text into chunks that respect sentence boundaries."""
    if max_length <= 0 or len(text) <= max_length:
        return [text]

    split_marks = set(punctuation_marks)

    chunks: List[str] = []
    remainder = text

    if node:
        send_log(
            node,
            "DEBUG",
            f"Splitting long text (len={len(text)}) with max={max_length} chars",
            log_level,
        )

    while remainder:
        if len(remainder) <= max_length:
            chunks.append(remainder)
            break

        # Find best split point
        split_idx = find_split_index(remainder, max_length, split_marks)
        if split_idx == -1:
            # No good split point found, force split at max_length
            split_idx = max_length

        chunk = remainder[:split_idx].strip()
        if chunk:
            chunks.append(chunk)
            if node:
                send_log(
                    node,
                    "DEBUG",
                    f"Created chunk: '{chunk[:30]}...' (len={len(chunk)})",
                    log_level,
                )

        remainder = remainder[split_idx:].lstrip()

    return chunks


def parse_markdown(file_path):
    """Parse markdown and extract segments for supported characters."""
    segments: List[Tuple[str, str]] = []
    current_character: Optional[str] = None
    current_text: List[str] = []

    # Map markdown tag content to our channel identifiers
    character_aliases: Dict[str, str] = {
        "大牛": "daniu",
        "一帆": "yifan",
        "博宇": "boyu",
        "Boyu": "boyu",
        "boyu": "boyu",
    }

    def finalize_segment() -> None:
        nonlocal current_character, current_text
        if current_character and current_text:
            combined_text = " ".join(current_text).strip()
            if combined_text:
                segments.append((current_character, combined_text))
        current_text = []

    with open(file_path, 'r', encoding='utf-8') as f:
        for raw_line in f:
            line = raw_line.strip()

            # Skip empty lines and headers
            if not line or line.startswith('#'):
                continue

            # Detect character tags inside the line
            match = re.search(r'【([^】]+)】', line)
            if match:
                finalize_segment()

                tag_content = match.group(1).strip()
                character = character_aliases.get(tag_content)
                if not character:
                    # Unknown character tag; skip this section entirely
                    current_character = None
                    current_text = []
                    continue

                # Extract the text following the tag, removing markdown markers
                remainder = line.split('】', 1)[1] if '】' in line else ''
                remainder = remainder.lstrip('*').strip()

                current_character = character
                current_text = [remainder] if remainder else []
                continue

            if current_character:
                # Continue accumulating text for current character
                clean_line = line.strip('*').strip()
                if clean_line:
                    current_text.append(clean_line)

        finalize_segment()

    return segments


@run_agent
def run(agent: MofaAgent):
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-file', required=True, help='Path to markdown script file')
    args = parser.parse_args()

    log_level = os.getenv("LOG_LEVEL", "INFO")

    # Text segmentation configuration
    max_segment_duration = float(os.getenv("MAX_SEGMENT_DURATION", "10.0"))
    chars_per_second = float(os.getenv("TTS_CHARS_PER_SECOND", "4.5"))
    max_segment_length = int(max_segment_duration * chars_per_second)

    # Punctuation marks for intelligent splitting
    punctuation_marks = os.getenv("PUNCTUATION_MARKS", "。！？.!?，,、；;：:")

    send_log(agent, "INFO", "Script Segmenter started", log_level)
    send_log(
        agent,
        "INFO",
        f"Text segmentation config: max_duration={max_segment_duration}s, "
        f"chars_per_second={chars_per_second}, max_length={max_segment_length} chars, "
        f"punctuation='{punctuation_marks}'",
        log_level
    )

    # Parse script
    try:
        segments = parse_markdown(args.input_file)
        send_log(agent, "INFO", f"Loaded {len(segments)} segments from {args.input_file}", log_level)
    except Exception as e:
        send_log(agent, "ERROR", f"Failed to parse script: {e}", log_level)
        return

    if not segments:
        send_log(agent, "ERROR", "No segments found in script!", log_level)
        return

    # Apply text segmentation to split long segments
    expanded_segments = []
    for character, text in segments:
        # Check if text is too long
        if len(text) > max_segment_length:
            send_log(
                agent,
                "INFO",
                f"Splitting long segment for {character} (len={len(text)}) into chunks",
                log_level
            )
            # Split into smaller chunks
            chunks = split_long_text(text, max_segment_length, punctuation_marks, agent, log_level)
            send_log(
                agent,
                "INFO",
                f"Split into {len(chunks)} chunks: {[len(c) for c in chunks]}",
                log_level
            )
            for chunk in chunks:
                expanded_segments.append((character, chunk))
        else:
            expanded_segments.append((character, text))

    segments = expanded_segments
    send_log(
        agent,
        "INFO",
        f"After text segmentation: {len(segments)} total segments to process",
        log_level
    )

    # State
    current_index = 0
    waiting_for = None

    # Send first segment BEFORE entering event loop
    character, text = segments[0]
    send_log(agent, "INFO", f"Sending segment 1/{len(segments)} to {character}: '{text[:50]}...' (len={len(text)})", log_level)
    agent.node.send_output(f'{character}_text', pa.array([text]))
    waiting_for = character
    current_index = 1
    send_log(agent, "INFO", f"Waiting for {waiting_for}_segment_complete...", log_level)

    # Event loop - wait for segment_complete and send next segments
    send_log(agent, "INFO", "Entering event loop", log_level)
    for event in agent.node:
        send_log(agent, "DEBUG", f"Received event type: {event['type']}", log_level)

        if event["type"] == "INPUT":
            event_id = event["id"]
            send_log(agent, "DEBUG", f"Received INPUT event: {event_id}", log_level)

            if event_id == f"{waiting_for}_segment_complete":
                send_log(agent, "INFO", f"Received {event_id}", log_level)

                if current_index < len(segments):
                    character, text = segments[current_index]
                    send_log(agent, "INFO", f"Sending segment {current_index+1}/{len(segments)} to {character}: '{text[:50]}...' (len={len(text)})", log_level)
                    agent.node.send_output(f'{character}_text', pa.array([text]))
                    waiting_for = character
                    current_index += 1
                else:
                    send_log(agent, "INFO", "All segments processed. Sending script_complete.", log_level)
                    agent.node.send_output('script_complete', pa.array([b'']))
                    send_log(agent, "INFO", "Script segmenter finished", log_level)
                    break

        elif event["type"] == "STOP":
            send_log(agent, "INFO", "Received STOP event", log_level)
            break

    send_log(agent, "INFO", "Event loop exited", log_level)


def main():
    agent = MofaAgent(agent_name="script-segmenter", is_write_log=True)
    run(agent=agent)


if __name__ == "__main__":
    main()



import json
import time
import pandas as pd
import os
import yaml
def get_log(dbname="logs.db", table="chat_completions"):
    import sqlite3

    con = sqlite3.connect(dbname)
    query = f"SELECT * from {table}"
    cursor = con.execute(query)
    rows = cursor.fetchall()
    column_names = [description[0] for description in cursor.description]
    data = [dict(zip(column_names, row)) for row in rows]
    con.close()
    return data
def str_to_dict(s):
    return json.loads(s)
def get_local_timestamp():
    return int(time.time())

def create_sqlite_db_name():
    return str(get_local_timestamp()) + '_logs.db'
def remove_sqlite_db_file(dbname="logs.db"):
    if os.path.exists(dbname):
        # 删除文件
        os.remove(dbname)


def load_log_token(logging_session_id:int,dbname="logs.db", table="chat_completions"):
    log_data = get_log(dbname=dbname,table=table)
    log_data_df = pd.DataFrame(log_data)

    log_data_df["total_tokens"] = log_data_df.apply(
        lambda row: str_to_dict(row["response"])["usage"]["total_tokens"], axis=1
    )

    log_data_df["request"] = log_data_df.apply(lambda row: str_to_dict(row["request"])["messages"][0]["content"], axis=1)

    log_data_df["response"] = log_data_df.apply(
        lambda row: str_to_dict(row["response"])["choices"][0]["message"]["content"], axis=1
    )

    total_tokens = log_data_df["total_tokens"].sum()

    # Sum total cost for all sessions
    total_cost = log_data_df["cost"].sum()

    session_tokens = log_data_df[log_data_df["session_id"] == logging_session_id]["total_tokens"].sum()
    session_cost = log_data_df[log_data_df["session_id"] == logging_session_id]["cost"].sum()

    return {'total_tokens':str(total_tokens),'total_cost': str(round(total_cost, 4)),
            'session_id':logging_session_id,}






manual_seed = 1313
model_source = 'bert-base-chinese'
window_size = 32
num_workers = 2
use_mask = True
use_conditional = True
param_conditional = {
    'bias': True,
    'char-linear': True,
    'pos-linear': False,
    'char+pos-second': True,
}

batch_size = 256
use_pos = True
param_pos = {
    'weight': 0.1,
    'pos_joint_training': True,
}



import os
import re

import cn2an
from pypinyin import lazy_pinyin, Style

from moyoyo_tts.text.symbols import punctuation
from moyoyo_tts.text.tone_sandhi import ToneSandhi
from moyoyo_tts.text.zh_normalization.text_normlization import TextNormalizer

normalizer = lambda x: cn2an.transform(x, "an2cn")

current_file_path = os.path.dirname(__file__)
pinyin_to_symbol_map = {
    line.split("\t")[0]: line.strip().split("\t")[1]
    for line in open(os.path.join(current_file_path, "opencpop-strict.txt")).readlines()
}

import jieba_fast.posseg as psg


rep_map = {
    "：": ",",
    "；": ",",
    "，": ",",
    "。": ".",
    "！": "!",
    "？": "?",
    "\n": ".",
    "·": ",",
    "、": ",",
    "...": "…",
    "$": ".",
    "/": ",",
    "—": "-",
    "~": "…",
    "～":"…",
}

tone_modifier = ToneSandhi()


def replace_punctuation(text):
    text = text.replace("嗯", "恩").replace("呣", "母")
    pattern = re.compile("|".join(re.escape(p) for p in rep_map.keys()))

    replaced_text = pattern.sub(lambda x: rep_map[x.group()], text)

    replaced_text = re.sub(
        r"[^\u4e00-\u9fa5" + "".join(punctuation) + r"]+", "", replaced_text
    )

    return replaced_text


def replace_punctuation_with_en(text):
    text = text.replace("嗯", "恩").replace("呣", "母")
    pattern = re.compile("|".join(re.escape(p) for p in rep_map.keys()))

    replaced_text = pattern.sub(lambda x: rep_map[x.group()], text)

    replaced_text = re.sub(
        r"[^\u4e00-\u9fa5A-Za-z" + "".join(punctuation) + r"]+", "", replaced_text
    )

    return replaced_text


def replace_consecutive_punctuation(text):
    punctuations = ''.join(re.escape(p) for p in punctuation)
    pattern = f'([{punctuations}])([{punctuations}])+'
    result = re.sub(pattern, r'\1', text)
    return result


def g2p(text):
    pattern = r"(?<=[{0}])\s*".format("".join(punctuation))
    sentences = [i for i in re.split(pattern, text) if i.strip() != ""]
    phones, word2ph = _g2p(sentences)
    return phones, word2ph


def _get_initials_finals(word):
    initials = []
    finals = []
    orig_initials = lazy_pinyin(word, neutral_tone_with_five=True, style=Style.INITIALS)
    orig_finals = lazy_pinyin(
        word, neutral_tone_with_five=True, style=Style.FINALS_TONE3
    )
    for c, v in zip(orig_initials, orig_finals):
        initials.append(c)
        finals.append(v)
    return initials, finals


def _g2p(segments):
    phones_list = []
    word2ph = []
    for seg in segments:
        pinyins = []
        # Replace all English words in the sentence
        seg = re.sub("[a-zA-Z]+", "", seg)
        seg_cut = psg.lcut(seg)
        initials = []
        finals = []
        seg_cut = tone_modifier.pre_merge_for_modify(seg_cut)
        for word, pos in seg_cut:
            if pos == "eng":
                continue
            sub_initials, sub_finals = _get_initials_finals(word)
            sub_finals = tone_modifier.modified_tone(word, pos, sub_finals)
            initials.append(sub_initials)
            finals.append(sub_finals)

            # assert len(sub_initials) == len(sub_finals) == len(word)
        initials = sum(initials, [])
        finals = sum(finals, [])
        #
        for c, v in zip(initials, finals):
            raw_pinyin = c + v
            # NOTE: post process for pypinyin outputs
            # we discriminate i, ii and iii
            if c == v:
                assert c in punctuation
                phone = [c]
                word2ph.append(1)
            else:
                v_without_tone = v[:-1]
                tone = v[-1]

                pinyin = c + v_without_tone
                assert tone in "12345"

                if c:
                    # 多音节
                    v_rep_map = {
                        "uei": "ui",
                        "iou": "iu",
                        "uen": "un",
                    }
                    if v_without_tone in v_rep_map.keys():
                        pinyin = c + v_rep_map[v_without_tone]
                else:
                    # 单音节
                    pinyin_rep_map = {
                        "ing": "ying",
                        "i": "yi",
                        "in": "yin",
                        "u": "wu",
                    }
                    if pinyin in pinyin_rep_map.keys():
                        pinyin = pinyin_rep_map[pinyin]
                    else:
                        single_rep_map = {
                            "v": "yu",
                            "e": "e",
                            "i": "y",
                            "u": "w",
                        }
                        if pinyin[0] in single_rep_map.keys():
                            pinyin = single_rep_map[pinyin[0]] + pinyin[1:]

                assert pinyin in pinyin_to_symbol_map.keys(), (pinyin, seg, raw_pinyin)
                new_c, new_v = pinyin_to_symbol_map[pinyin].split(" ")
                new_v = new_v + tone
                phone = [new_c, new_v]
                word2ph.append(len(phone))

            phones_list += phone
    return phones_list, word2ph


def text_normalize(text):
    # https://github.com/PaddlePaddle/PaddleSpeech/tree/develop/paddlespeech/t2s/frontend/zh_normalization
    tx = TextNormalizer()
    sentences = tx.normalize(text)
    dest_text = ""
    for sentence in sentences:
        dest_text += replace_punctuation(sentence)

    # 避免重复标点引起的参考泄露
    dest_text = replace_consecutive_punctuation(dest_text)
    return dest_text


# 不排除英文的文本格式化
def mix_text_normalize(text):
    # https://github.com/PaddlePaddle/PaddleSpeech/tree/develop/paddlespeech/t2s/frontend/zh_normalization
    tx = TextNormalizer()
    sentences = tx.normalize(text)
    dest_text = ""
    for sentence in sentences:
        dest_text += replace_punctuation_with_en(sentence)

    # 避免重复标点引起的参考泄露
    dest_text = replace_consecutive_punctuation(dest_text)
    return dest_text


if __name__ == "__main__":
    text = "啊——但是《原神》是由,米哈\游自主，研发的一款全.新开放世界.冒险游戏"
    text = "呣呣呣～就是…大人的鼹鼠党吧？"
    text = "你好"
    text = text_normalize(text)
    print(g2p(text))


# # 示例用法
# text = "这是一个示例文本：,你好！这是一个测试..."
# print(g2p_paddle(text))  # 输出: 这是一个示例文本你好这是一个测试



import argparse
import json
import os
import ast
import sys

import click
import pyarrow as pa
from dora import Node
from mofa.utils.install_pkg.load_task_weaver_result import extract_important_content
RUNNER_CI = True if os.getenv("CI") == "true" else False




def clean_string(input_string:str):
    return input_string.encode('utf-8', 'replace').decode('utf-8')
def send_task_and_receive_data(node):
    TIMEOUT = 300
    while True:
        data = input(
            " Send Your Task :  ",
        )
        node.send_output("data", pa.array([clean_string(data)]))
        event = node.next(timeout=TIMEOUT)
        if event is not None:
            while True:
                if event is not None:
                    node_results = json.loads(event['value'].to_pylist()[0])
                    results = node_results.get('node_results')
                    is_dataflow_end = node_results.get('dataflow_status', False)
                    step_name = node_results.get('step_name', '')
                    click.echo(f'-------------{step_name}---------------')
                    click.echo(f"{results} ", )
                    click.echo(f'---------------------------------------')
                    sys.stdout.flush()
                    if is_dataflow_end ==True or is_dataflow_end == 'true' or is_dataflow_end == 'True':
                        break
                    event = node.next(timeout=TIMEOUT)
def main():

    # Handle dynamic nodes, ask for the name of the node in the dataflow, and the same values as the ENV variables.
    parser = argparse.ArgumentParser(description="Simple arrow sender")

    parser.add_argument(
        "--name",
        type=str,
        required=False,
        help="The name of the node in the dataflow.",
        default="terminal-input",
    )
    parser.add_argument(
        "--data",
        type=str,
        required=False,
        help="Arrow Data as string.",
        default=None,
    )

    args = parser.parse_args()

    data = os.getenv("DATA", args.data)

    node = Node(
        args.name
    )  # provide the name to connect to the dataflow if dynamic node

    if data is None and os.getenv("DORA_NODE_CONFIG") is None:
        send_task_and_receive_data(node)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Voice output agent for the MoFA podcast generator flow.
Concatenates incoming audio streams, adds speaker-switch silences,
and persists the final WAV.
"""

import argparse
import json
import os
import random
import time
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
import pyarrow as pa
from scipy.io import wavfile

from mofa.agent_build.base.base_agent import MofaAgent, run_agent


def send_log(agent: MofaAgent, level: str, message: str, config_level: str = "INFO") -> None:
    """Send structured logs through the node output."""
    log_levels = {"DEBUG": 10, "INFO": 20, "WARNING": 30, "ERROR": 40}
    if log_levels.get(level, 0) < log_levels.get(config_level, 20):
        return

    formatted = f"[{level}] {message}"
    log_payload = {
        "node": agent.agent_name,
        "level": level,
        "message": formatted,
        "timestamp": time.time(),
    }
    agent.node.send_output("log", pa.array([json.dumps(log_payload)]))


def _ensure_output_directory(output_file: Path, agent: MofaAgent, log_level: str) -> None:
    """Create the output directory when needed."""
    try:
        output_file.parent.mkdir(parents=True, exist_ok=True)
    except Exception as exc:
        send_log(
            agent,
            "ERROR",
            f"Failed to create output directory '{output_file.parent}': {exc}",
            log_level,
        )
        raise


@run_agent
def run(agent: MofaAgent, output_file: Path, log_level: str) -> None:
    sample_rate = int(os.getenv("VOICE_OUTPUT_SAMPLE_RATE", "32000"))
    silence_min = float(os.getenv("VOICE_OUTPUT_SILENCE_MIN", "0.3"))
    silence_max = float(os.getenv("VOICE_OUTPUT_SILENCE_MAX", "1.2"))

    speakers: Dict[str, str] = {
        "daniu": "大牛",
        "yifan": "一帆",
        "boyu": "博宇",
    }

    audio_suffix = "_audio"
    segment_complete_suffix = "_segment_complete"

    audio_buffer: List[np.ndarray] = []
    last_speaker: Optional[str] = None
    segment_count = 0

    _ensure_output_directory(output_file, agent, log_level)
    send_log(agent, "INFO", f"Voice output initialized → {output_file}", log_level)
    send_log(
        agent,
        "INFO",
        f"Sample rate {sample_rate} Hz, silence window {silence_min}-{silence_max}s",
        log_level,
    )

    def process_audio_event(speaker_key: str, event) -> None:
        nonlocal last_speaker, segment_count

        speaker_name = speakers[speaker_key]
        metadata = event.get("metadata", {}) or {}
        fragment_num = metadata.get("fragment_num")
        is_segment_start = fragment_num == 1 if fragment_num is not None else True

        if is_segment_start and last_speaker is not None and last_speaker != speaker_key:
            silence_duration = random.uniform(silence_min, silence_max)
            silence_samples = int(sample_rate * silence_duration)
            silence = np.zeros(silence_samples, dtype=np.int16)
            audio_buffer.append(silence)
            previous_name = speakers.get(last_speaker, last_speaker)
            send_log(
                agent,
                "INFO",
                f"Added {silence_duration:.2f}s silence ({previous_name} → {speaker_name})",
                log_level,
            )

        if is_segment_start:
            last_speaker = speaker_key

        raw_value = event.get("value")
        if not raw_value:
            send_log(agent, "WARNING", f"No audio payload received from {speaker_name}", log_level)
            return

        audio_data = raw_value[0].as_py()
        if not isinstance(audio_data, np.ndarray):
            audio_data = np.array(audio_data, dtype=np.float32)

        original_dtype = audio_data.dtype
        if audio_data.dtype in (np.float32, np.float64):
            audio_data = (audio_data * 32767).astype(np.int16)
        elif audio_data.dtype != np.int16:
            send_log(
                agent,
                "WARNING",
                f"Unexpected dtype {original_dtype} from {speaker_name}, casting to int16",
                log_level,
            )
            audio_data = audio_data.astype(np.int16)

        audio_buffer.append(audio_data)
        segment_count += 1
        fragment_label = fragment_num if fragment_num is not None else "?"
        send_log(
            agent,
            "INFO",
            f"✓ Audio from {speaker_name} (fragment #{fragment_label}, {len(audio_data)} samples)",
            log_level,
        )

    for event in agent.node:
        if event["type"] == "INPUT":
            event_id = event["id"]

            if event_id.endswith(audio_suffix):
                speaker_key = event_id[: -len(audio_suffix)]
                if speaker_key in speakers:
                    process_audio_event(speaker_key, event)

            elif event_id.endswith(segment_complete_suffix):
                speaker_key = event_id[: -len(segment_complete_suffix)]
                if speaker_key in speakers:
                    last_speaker = speaker_key
                    send_log(
                        agent,
                        "DEBUG",
                        f"{speakers[speaker_key]} segment complete, awaiting next speaker",
                        log_level,
                    )

            elif event_id == "script_complete":
                send_log(
                    agent,
                    "INFO",
                    f"Script complete after {segment_count} fragments. Concatenating…",
                    log_level,
                )

                if not audio_buffer:
                    send_log(agent, "ERROR", "No audio received; skipping file write.", log_level)
                else:
                    try:
                        for idx, array in enumerate(audio_buffer):
                            if array.dtype != np.int16:
                                audio_buffer[idx] = array.astype(np.int16)

                        final_audio = np.concatenate(audio_buffer)
                        wavfile.write(output_file, sample_rate, final_audio)

                        try:
                            os.sync()
                        except AttributeError:
                            pass

                        duration = len(final_audio) / sample_rate
                        file_size = output_file.stat().st_size
                        send_log(
                            agent,
                            "INFO",
                            f"Podcast saved → {output_file} ({duration:.2f}s, {file_size} bytes)",
                            log_level,
                        )
                    except Exception as exc:
                        send_log(agent, "ERROR", f"Failed to write audio: {exc}", log_level)

                time.sleep(0.1)
                send_log(agent, "INFO", "Voice output agent exiting", log_level)
                break

        elif event["type"] == "STOP":
            send_log(agent, "INFO", "Received STOP event", log_level)
            break


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="MoFA podcast voice output agent")
    parser.add_argument(
        "--output-file",
        default=os.getenv("PODCAST_OUTPUT_FILE", "out/podcast_output.wav"),
        help="Path to the output WAV file",
    )
    return parser.parse_args()


def main() -> None:
    args = _parse_args()
    output_path = Path(args.output_file).expanduser().resolve()
    log_level = os.getenv("LOG_LEVEL", "INFO")

    agent = MofaAgent(agent_name="voice-output", is_write_log=True)
    run(agent=agent, output_file=output_path, log_level=log_level)


if __name__ == "__main__":
    main()



import json
import os

import locale

I18N_JSON_DIR : os.PathLike = os.path.join(os.path.dirname(os.path.relpath(__file__)), 'locale')

def load_language_list(language):
    with open(os.path.join(I18N_JSON_DIR, f"{language}.json"), "r", encoding="utf-8") as f:
        language_list = json.load(f)
    return language_list

def scan_language_list():
    language_list = []
    for name in os.listdir(I18N_JSON_DIR):
        if name.endswith(".json"):language_list.append(name.split('.')[0])
    return language_list

class I18nAuto:
    def __init__(self, language=None):
        if language in ["Auto", None]:
            language = locale.getdefaultlocale()[0]  
            # getlocale can't identify the system's language ((None, None))
        if not os.path.exists(os.path.join(I18N_JSON_DIR, f"{language}.json")):
            language = "en_US"
        self.language = language
        self.language_map = load_language_list(language)

    def __call__(self, key):
        return self.language_map.get(key, key)

    def __repr__(self):
        return "Use Language: " + self.language

if __name__ == "__main__":
    i18n = I18nAuto(language='en_US')
    print(i18n)


# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/text_processing/symbols.py
# reference: https://github.com/lifeiteng/vall-e
PAD = "_"
PUNCTUATION = ';:,.!?¡¿—…"«»“” '
LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
IPA_LETTERS = "ɑɐɒæɓʙβɔɕçɗɖðʤəɘɚɛɜɝɞɟʄɡɠɢʛɦɧħɥʜɨɪʝɭɬɫɮʟɱɯɰŋɳɲɴøɵɸθœɶʘɹɺɾɻʀʁɽʂʃʈʧʉʊʋⱱʌɣɤʍχʎʏʑʐʒʔʡʕʢǀǁǂǃˈˌːˑʼʴʰʱʲʷˠˤ˞↓↑→↗↘'̩'ᵻ"
SYMBOLS = [PAD] + list(PUNCTUATION) + list(LETTERS) + list(IPA_LETTERS)
SPACE_ID = SYMBOLS.index(" ")
SYMBOL_TO_ID = {s: i for i, s in enumerate(SYMBOLS)}
ID_TO_SYMBOL = {i: s for i, s in enumerate(SYMBOLS)}



# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/data/bucket_sampler.py
# reference: https://github.com/lifeiteng/vall-e
import itertools
import math
import random
from random import shuffle
from typing import Iterator
from typing import Optional
from typing import TypeVar

import torch
import torch.distributed as dist
from torch.utils.data import Dataset
from torch.utils.data import Sampler

__all__ = [
    "DistributedBucketSampler",
]

T_co = TypeVar("T_co", covariant=True)


class DistributedBucketSampler(Sampler[T_co]):
    r"""
    sort the dataset wrt. input length
    divide samples into buckets
    sort within buckets
    divide buckets into batches
    sort batches
    """

    def __init__(
        self,
        dataset: Dataset,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
        batch_size: int = 32,
    ) -> None:
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size() if torch.cuda.is_available() else 1
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank() if torch.cuda.is_available() else 0
            if torch.cuda.is_available():
                torch.cuda.set_device(rank)
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                "Invalid rank {}, rank should be in the interval"
                " [0, {}]".format(rank, num_replicas - 1)
            )
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last
        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        if (
            self.drop_last and len(self.dataset) % self.num_replicas != 0
        ):  # type: ignore[arg-type]
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = math.ceil(
                (len(self.dataset) - self.num_replicas)
                / self.num_replicas  # type: ignore[arg-type]
            )
        else:
            self.num_samples = math.ceil(
                len(self.dataset) / self.num_replicas
            )  # type: ignore[arg-type]
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        self.seed = seed
        self.batch_size = batch_size
        self.id_with_length = self._get_sample_lengths()
        self.id_buckets = self.make_buckets(bucket_width=2.0)

    def _get_sample_lengths(self):
        id_with_lengths = []
        for i in range(len(self.dataset)):
            id_with_lengths.append((i, self.dataset.get_sample_length(i)))
        id_with_lengths.sort(key=lambda x: x[1])
        return id_with_lengths

    def make_buckets(self, bucket_width: float = 2.0):
        buckets = []
        cur = []
        max_sec = bucket_width
        for id, sec in self.id_with_length:
            if sec < max_sec:
                cur.append(id)
            else:
                buckets.append(cur)
                cur = [id]
                max_sec += bucket_width
        if len(cur) > 0:
            buckets.append(cur)
        return buckets

    def __iter__(self) -> Iterator[T_co]:
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            random.seed(self.epoch + self.seed)
            shuffled_bucket = []
            for buc in self.id_buckets:
                buc_copy = buc.copy()
                shuffle(buc_copy)
                shuffled_bucket.append(buc_copy)
            grouped_batch_size = self.batch_size * self.num_replicas
            shuffled_bucket = list(itertools.chain(*shuffled_bucket))
            n_batch = int(math.ceil(len(shuffled_bucket) / grouped_batch_size))
            batches = [
                shuffled_bucket[b * grouped_batch_size : (b + 1) * grouped_batch_size]
                for b in range(n_batch)
            ]
            shuffle(batches)
            indices = list(itertools.chain(*batches))
        else:
            # type: ignore[arg-type]
            indices = list(range(len(self.dataset)))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self) -> int:
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch



import os
import glob
import sys
import argparse
import logging
import json
import subprocess
import traceback

import librosa
import numpy as np
from scipy.io.wavfile import read
import torch
import logging

logging.getLogger("numba").setLevel(logging.ERROR)
logging.getLogger("matplotlib").setLevel(logging.ERROR)

MATPLOTLIB_FLAG = False

logging.basicConfig(stream=sys.stdout, level=logging.ERROR)
logger = logging


def load_checkpoint(checkpoint_path, model, optimizer=None, skip_optimizer=False):
    assert os.path.isfile(checkpoint_path)
    checkpoint_dict = torch.load(checkpoint_path, map_location="cpu")
    iteration = checkpoint_dict["iteration"]
    learning_rate = checkpoint_dict["learning_rate"]
    if (
        optimizer is not None
        and not skip_optimizer
        and checkpoint_dict["optimizer"] is not None
    ):
        optimizer.load_state_dict(checkpoint_dict["optimizer"])
    saved_state_dict = checkpoint_dict["model"]
    if hasattr(model, "module"):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    new_state_dict = {}
    for k, v in state_dict.items():
        try:
            # assert "quantizer" not in k
            # print("load", k)
            new_state_dict[k] = saved_state_dict[k]
            assert saved_state_dict[k].shape == v.shape, (
                saved_state_dict[k].shape,
                v.shape,
            )
        except:
            traceback.print_exc()
            print(
                "error, %s is not in the checkpoint" % k
            )  # shape不对也会，比如text_embedding当cleaner修改时
            new_state_dict[k] = v
    if hasattr(model, "module"):
        model.module.load_state_dict(new_state_dict)
    else:
        model.load_state_dict(new_state_dict)
    print("load ")
    logger.info(
        "Loaded checkpoint '{}' (iteration {})".format(checkpoint_path, iteration)
    )
    return model, optimizer, learning_rate, iteration

from time import time as ttime
import shutil
def my_save(fea,path):#####fix issue: torch.save doesn't support chinese path
    dir=os.path.dirname(path)
    name=os.path.basename(path)
    tmp_path="%s.pth"%(ttime())
    torch.save(fea,tmp_path)
    shutil.move(tmp_path,"%s/%s"%(dir,name))

def save_checkpoint(model, optimizer, learning_rate, iteration, checkpoint_path):
    logger.info(
        "Saving model and optimizer state at iteration {} to {}".format(
            iteration, checkpoint_path
        )
    )
    if hasattr(model, "module"):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    # torch.save(
    my_save(
        {
            "model": state_dict,
            "iteration": iteration,
            "optimizer": optimizer.state_dict(),
            "learning_rate": learning_rate,
        },
        checkpoint_path,
    )


def summarize(
    writer,
    global_step,
    scalars={},
    histograms={},
    images={},
    audios={},
    audio_sampling_rate=22050,
):
    for k, v in scalars.items():
        writer.add_scalar(k, v, global_step)
    for k, v in histograms.items():
        writer.add_histogram(k, v, global_step)
    for k, v in images.items():
        writer.add_image(k, v, global_step, dataformats="HWC")
    for k, v in audios.items():
        writer.add_audio(k, v, global_step, audio_sampling_rate)


def latest_checkpoint_path(dir_path, regex="G_*.pth"):
    f_list = glob.glob(os.path.join(dir_path, regex))
    f_list.sort(key=lambda f: int("".join(filter(str.isdigit, f))))
    x = f_list[-1]
    print(x)
    return x


def plot_spectrogram_to_numpy(spectrogram):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib

        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger("matplotlib")
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(10, 2))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower", interpolation="none")
    plt.colorbar(im, ax=ax)
    plt.xlabel("Frames")
    plt.ylabel("Channels")
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close()
    return data


def plot_alignment_to_numpy(alignment, info=None):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib

        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger("matplotlib")
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(6, 4))
    im = ax.imshow(
        alignment.transpose(), aspect="auto", origin="lower", interpolation="none"
    )
    fig.colorbar(im, ax=ax)
    xlabel = "Decoder timestep"
    if info is not None:
        xlabel += "\n\n" + info
    plt.xlabel(xlabel)
    plt.ylabel("Encoder timestep")
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close()
    return data


def load_wav_to_torch(full_path):
    data, sampling_rate = librosa.load(full_path, sr=None)
    return torch.FloatTensor(data), sampling_rate


def load_filepaths_and_text(filename, split="|"):
    with open(filename, encoding="utf-8") as f:
        filepaths_and_text = [line.strip().split(split) for line in f]
    return filepaths_and_text


def get_hparams(init=True, stage=1):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default="./configs/s2.json",
        help="JSON file for configuration",
    )
    parser.add_argument(
        "-p", "--pretrain", type=str, required=False, default=None, help="pretrain dir"
    )
    parser.add_argument(
        "-rs",
        "--resume_step",
        type=int,
        required=False,
        default=None,
        help="resume step",
    )
    # parser.add_argument('-e', '--exp_dir', type=str, required=False,default=None,help='experiment directory')
    # parser.add_argument('-g', '--pretrained_s2G', type=str, required=False,default=None,help='pretrained sovits gererator weights')
    # parser.add_argument('-d', '--pretrained_s2D', type=str, required=False,default=None,help='pretrained sovits discriminator weights')

    args = parser.parse_args()

    config_path = args.config
    with open(config_path, "r") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.pretrain = args.pretrain
    hparams.resume_step = args.resume_step
    # hparams.data.exp_dir = args.exp_dir
    if stage == 1:
        model_dir = hparams.s1_ckpt_dir
    else:
        model_dir = hparams.s2_ckpt_dir
    config_save_path = os.path.join(model_dir, "config.json")

    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    with open(config_save_path, "w") as f:
        f.write(data)
    return hparams


def clean_checkpoints(path_to_models="logs/44k/", n_ckpts_to_keep=2, sort_by_time=True):
    """Freeing up space by deleting saved ckpts

    Arguments:
    path_to_models    --  Path to the model directory
    n_ckpts_to_keep   --  Number of ckpts to keep, excluding G_0.pth and D_0.pth
    sort_by_time      --  True -> chronologically delete ckpts
                          False -> lexicographically delete ckpts
    """
    import re

    ckpts_files = [
        f
        for f in os.listdir(path_to_models)
        if os.path.isfile(os.path.join(path_to_models, f))
    ]
    name_key = lambda _f: int(re.compile("._(\d+)\.pth").match(_f).group(1))
    time_key = lambda _f: os.path.getmtime(os.path.join(path_to_models, _f))
    sort_key = time_key if sort_by_time else name_key
    x_sorted = lambda _x: sorted(
        [f for f in ckpts_files if f.startswith(_x) and not f.endswith("_0.pth")],
        key=sort_key,
    )
    to_del = [
        os.path.join(path_to_models, fn)
        for fn in (x_sorted("G")[:-n_ckpts_to_keep] + x_sorted("D")[:-n_ckpts_to_keep])
    ]
    del_info = lambda fn: logger.info(f".. Free up space by deleting ckpt {fn}")
    del_routine = lambda x: [os.remove(x), del_info(x)]
    rs = [del_routine(fn) for fn in to_del]


def get_hparams_from_dir(model_dir):
    config_save_path = os.path.join(model_dir, "config.json")
    with open(config_save_path, "r") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.model_dir = model_dir
    return hparams


def get_hparams_from_file(config_path):
    with open(config_path, "r") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    return hparams


def check_git_hash(model_dir):
    source_dir = os.path.dirname(os.path.realpath(__file__))
    if not os.path.exists(os.path.join(source_dir, ".git")):
        logger.warn(
            "{} is not a git repository, therefore hash value comparison will be ignored.".format(
                source_dir
            )
        )
        return

    cur_hash = subprocess.getoutput("git rev-parse HEAD")

    path = os.path.join(model_dir, "githash")
    if os.path.exists(path):
        saved_hash = open(path).read()
        if saved_hash != cur_hash:
            logger.warn(
                "git hash values are different. {}(saved) != {}(current)".format(
                    saved_hash[:8], cur_hash[:8]
                )
            )
    else:
        open(path, "w").write(cur_hash)


def get_logger(model_dir, filename="train.log"):
    global logger
    logger = logging.getLogger(os.path.basename(model_dir))
    logger.setLevel(logging.ERROR)

    formatter = logging.Formatter("%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    h = logging.FileHandler(os.path.join(model_dir, filename))
    h.setLevel(logging.ERROR)
    h.setFormatter(formatter)
    logger.addHandler(h)
    return logger


class HParams:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if type(v) == dict:
                v = HParams(**v)
            self[k] = v

    def keys(self):
        return self.__dict__.keys()

    def items(self):
        return self.__dict__.items()

    def values(self):
        return self.__dict__.values()

    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __contains__(self, key):
        return key in self.__dict__

    def __repr__(self):
        return self.__dict__.__repr__()


if __name__ == "__main__":
    print(
        load_wav_to_torch(
            "/home/fish/wenetspeech/dataset_vq/Y0000022499_wHFSeHEx9CM/S00261.flac"
        )
    )



import math
import torch
from torch import nn
from torch.nn import functional as F

from moyoyo_tts.module import commons

from typing import Optional

class LayerNorm(nn.Module):
    def __init__(self, channels, eps=1e-5):
        super().__init__()
        self.channels = channels
        self.eps = eps

        self.gamma = nn.Parameter(torch.ones(channels))
        self.beta = nn.Parameter(torch.zeros(channels))

    def forward(self, x):
        x = x.transpose(1, -1)
        x = F.layer_norm(x, (self.channels,), self.gamma, self.beta, self.eps)
        return x.transpose(1, -1)


@torch.jit.script
def fused_add_tanh_sigmoid_multiply(input_a, input_b, n_channels):
    n_channels_int = n_channels[0]
    in_act = input_a + input_b
    t_act = torch.tanh(in_act[:, :n_channels_int, :])
    s_act = torch.sigmoid(in_act[:, n_channels_int:, :])
    acts = t_act * s_act
    return acts


class Encoder(nn.Module):
    def __init__(
        self,
        hidden_channels,
        filter_channels,
        n_heads,
        n_layers,
        kernel_size=1,
        p_dropout=0.0,
        window_size=4,
        isflow=True,
        **kwargs
    ):
        super().__init__()
        self.hidden_channels = hidden_channels
        self.filter_channels = filter_channels
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.window_size = window_size
        # if isflow:
        #  cond_layer = torch.nn.Conv1d(256, 2*hidden_channels*n_layers, 1)
        #  self.cond_pre = torch.nn.Conv1d(hidden_channels, 2*hidden_channels, 1)
        #  self.cond_layer = weight_norm(cond_layer, name='weight')
        #  self.gin_channels = 256
        self.cond_layer_idx = self.n_layers
        self.spk_emb_linear = nn.Linear(256, self.hidden_channels)
        if "gin_channels" in kwargs:
            self.gin_channels = kwargs["gin_channels"]
            if self.gin_channels != 0:
                self.spk_emb_linear = nn.Linear(self.gin_channels, self.hidden_channels)
                # vits2 says 3rd block, so idx is 2 by default
                self.cond_layer_idx = (
                    kwargs["cond_layer_idx"] if "cond_layer_idx" in kwargs else 2
                )
                logging.debug(self.gin_channels, self.cond_layer_idx)
                assert (
                    self.cond_layer_idx < self.n_layers
                ), "cond_layer_idx should be less than n_layers"
        self.drop = nn.Dropout(p_dropout)
        self.attn_layers = nn.ModuleList()
        self.norm_layers_1 = nn.ModuleList()
        self.ffn_layers = nn.ModuleList()
        self.norm_layers_2 = nn.ModuleList()
        for i in range(self.n_layers):
            self.attn_layers.append(
                MultiHeadAttention(
                    hidden_channels,
                    hidden_channels,
                    n_heads,
                    p_dropout=p_dropout,
                    window_size=window_size,
                )
            )
            self.norm_layers_1.append(LayerNorm(hidden_channels))
            self.ffn_layers.append(
                FFN(
                    hidden_channels,
                    hidden_channels,
                    filter_channels,
                    kernel_size,
                    p_dropout=p_dropout,
                )
            )
            self.norm_layers_2.append(LayerNorm(hidden_channels))

    # def forward(self, x, x_mask, g=None):
    #     attn_mask = x_mask.unsqueeze(2) * x_mask.unsqueeze(-1)
    #     x = x * x_mask
    #     for i in range(self.n_layers):
    #         if i == self.cond_layer_idx and g is not None:
    #             g = self.spk_emb_linear(g.transpose(1, 2))
    #             g = g.transpose(1, 2)
    #             x = x + g
    #             x = x * x_mask
    #         y = self.attn_layers[i](x, x, attn_mask)
    #         y = self.drop(y)
    #         x = self.norm_layers_1[i](x + y)

    #         y = self.ffn_layers[i](x, x_mask)
    #         y = self.drop(y)
    #         x = self.norm_layers_2[i](x + y)
    #     x = x * x_mask
    #     return x

    def forward(self, x, x_mask):
        attn_mask = x_mask.unsqueeze(2) * x_mask.unsqueeze(-1)
        x = x * x_mask
        for attn_layers,norm_layers_1,ffn_layers,norm_layers_2 in zip(self.attn_layers,self.norm_layers_1,self.ffn_layers,self.norm_layers_2):
            y = attn_layers(x, x, attn_mask)
            y = self.drop(y)
            x = norm_layers_1(x + y)

            y = ffn_layers(x, x_mask)
            y = self.drop(y)
            x = norm_layers_2(x + y)
        x = x * x_mask
        return x


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        channels,
        out_channels,
        n_heads,
        p_dropout=0.0,
        window_size=None,
        heads_share=True,
        block_length=None,
        proximal_bias=False,
        proximal_init=False,
    ):
        super().__init__()
        assert channels % n_heads == 0

        self.channels = channels
        self.out_channels = out_channels
        self.n_heads = n_heads
        self.p_dropout = p_dropout
        self.window_size = window_size
        self.heads_share = heads_share
        self.block_length = block_length
        self.proximal_bias = proximal_bias
        self.proximal_init = proximal_init
        self.attn = None

        self.k_channels = channels // n_heads
        self.conv_q = nn.Conv1d(channels, channels, 1)
        self.conv_k = nn.Conv1d(channels, channels, 1)
        self.conv_v = nn.Conv1d(channels, channels, 1)
        self.conv_o = nn.Conv1d(channels, out_channels, 1)
        self.drop = nn.Dropout(p_dropout)

        if window_size is not None:
            n_heads_rel = 1 if heads_share else n_heads
            rel_stddev = self.k_channels**-0.5
            self.emb_rel_k = nn.Parameter(
                torch.randn(n_heads_rel, window_size * 2 + 1, self.k_channels)
                * rel_stddev
            )
            self.emb_rel_v = nn.Parameter(
                torch.randn(n_heads_rel, window_size * 2 + 1, self.k_channels)
                * rel_stddev
            )

        nn.init.xavier_uniform_(self.conv_q.weight)
        nn.init.xavier_uniform_(self.conv_k.weight)
        nn.init.xavier_uniform_(self.conv_v.weight)
        if proximal_init:
            with torch.no_grad():
                self.conv_k.weight.copy_(self.conv_q.weight)
                self.conv_k.bias.copy_(self.conv_q.bias)

    def forward(self, x, c, attn_mask:Optional[torch.Tensor]=None):
        q = self.conv_q(x)
        k = self.conv_k(c)
        v = self.conv_v(c)

        # x, self.attn = self.attention(q, k, v, mask=attn_mask)
        x, _ = self.attention(q, k, v, mask=attn_mask)

        x = self.conv_o(x)
        return x

    def attention(self, query, key, value, mask:Optional[torch.Tensor]=None):
        # reshape [b, d, t] -> [b, n_h, t, d_k]
        b, d, t_s, _ = (*key.size(), query.size(2))
        query = query.view(b, self.n_heads, self.k_channels, -1).transpose(2, 3)
        key = key.view(b, self.n_heads, self.k_channels, -1).transpose(2, 3)
        value = value.view(b, self.n_heads, self.k_channels, -1).transpose(2, 3)
        scores = torch.matmul(query / math.sqrt(self.k_channels), key.transpose(-2, -1))

        if self.window_size is not None:
            key_relative_embeddings = self._get_relative_embeddings(self.emb_rel_k, t_s)
            rel_logits = self._matmul_with_relative_keys(query / math.sqrt(self.k_channels), key_relative_embeddings)
            scores_local = self._relative_position_to_absolute_position(rel_logits)
            scores = scores + scores_local

        if mask is not None:
            scores = scores.masked_fill(mask == 0, -1e4)

        p_attn = F.softmax(scores, dim=-1)
        p_attn = self.drop(p_attn)
        output = torch.matmul(p_attn, value)

        if self.window_size is not None:
            relative_weights = self._absolute_position_to_relative_position(p_attn)
            value_relative_embeddings = self._get_relative_embeddings(self.emb_rel_v, t_s)
            output = output + self._matmul_with_relative_values(relative_weights, value_relative_embeddings)

        output = (output.transpose(2, 3).contiguous().view(b, d, -1))
        return output, p_attn

    def _matmul_with_relative_values(self, x, y):
        """
        x: [b, h, l, m]
        y: [h or 1, m, d]
        ret: [b, h, l, d]
        """
        ret = torch.matmul(x, y.unsqueeze(0))
        return ret

    def _matmul_with_relative_keys(self, x, y):
        """
        x: [b, h, l, d]
        y: [h or 1, m, d]
        ret: [b, h, l, m]
        """
        ret = torch.matmul(x, y.unsqueeze(0).transpose(-2, -1))
        return ret

    def _get_relative_embeddings(self, relative_embeddings, length):
        max_relative_position = 2 * self.window_size + 1
        # Pad first before slice to avoid using cond ops.
        pad_l = torch.zeros((1), dtype = torch.int64) + length - (self.window_size + 1)
        pad_s = torch.zeros((1), dtype = torch.int64) + (self.window_size + 1) - length
        pad_length = torch.max(pad_l, other=torch.zeros((1), dtype = torch.int64))
        slice_start_position = torch.max(pad_s, other=torch.zeros((1), dtype = torch.int64))

        slice_end_position = slice_start_position + 2 * length - 1
        padded_relative_embeddings = F.pad(
            relative_embeddings,
            commons.convert_pad_shape([[0, 0], [pad_length, pad_length], [0, 0]]),
        )
        used_relative_embeddings = padded_relative_embeddings[
            :, slice_start_position:slice_end_position
        ]
        return used_relative_embeddings

    def _relative_position_to_absolute_position(self, x):
        """
        x: [b, h, l, 2*l-1]
        ret: [b, h, l, l]
        """
        batch, heads, length, _ = x.size()
        # Concat columns of pad to shift from relative to absolute indexing.
        x = F.pad(x, commons.convert_pad_shape([[0, 0], [0, 0], [0, 0], [0, 1]]))

        # Concat extra elements so to add up to shape (len+1, 2*len-1).
        x_flat = x.view([batch, heads, length * 2 * length])
        x_flat = F.pad(
            x_flat, commons.convert_pad_shape([[0, 0], [0, 0], [0, length - 1]])
        )

        # Reshape and slice out the padded elements.
        x_final = x_flat.view([batch, heads, length + 1, 2 * length - 1])[
            :, :, :length, length - 1 :
        ]
        return x_final

    def _absolute_position_to_relative_position(self, x):
        """
        x: [b, h, l, l]
        ret: [b, h, l, 2*l-1]
        """
        batch, heads, length, _ = x.size()
        # padd along column
        x = F.pad(
            x, commons.convert_pad_shape([[0, 0], [0, 0], [0, 0], [0, length - 1]])
        )
        x_flat = x.view([batch, heads, length**2 + length * (length - 1)])
        # add 0's in the beginning that will skew the elements after reshape
        x_flat = F.pad(x_flat, commons.convert_pad_shape([[0, 0], [0, 0], [length, 0]]))
        x_final = x_flat.view([batch, heads, length, 2 * length])[:, :, :, 1:]
        return x_final

    def _attention_bias_proximal(self, length):
        """Bias for self-attention to encourage attention to close positions.
        Args:
          length: an integer scalar.
        Returns:
          a Tensor with shape [1, 1, length, length]
        """
        r = torch.arange(length, dtype=torch.float32)
        diff = torch.unsqueeze(r, 0) - torch.unsqueeze(r, 1)
        return torch.unsqueeze(torch.unsqueeze(-torch.log1p(torch.abs(diff)), 0), 0)


class FFN(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        filter_channels,
        kernel_size,
        p_dropout=0.0,
        activation="",
        causal=False,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.filter_channels = filter_channels
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.activation = activation
        self.causal = causal

        # 从上下文看这里一定是 False
        # if causal:
        #     self.padding = self._causal_padding
        # else:
        #     self.padding = self._same_padding

        self.conv_1 = nn.Conv1d(in_channels, filter_channels, kernel_size)
        self.conv_2 = nn.Conv1d(filter_channels, out_channels, kernel_size)
        self.drop = nn.Dropout(p_dropout)

    def forward(self, x, x_mask):
        x = self.conv_1(self.padding(x * x_mask))
        if self.activation == "gelu":
            x = x * torch.sigmoid(1.702 * x)
        else:
            x = torch.relu(x)
        x = self.drop(x)
        x = self.conv_2(self.padding(x * x_mask))
        return x * x_mask

    def padding(self, x):
        return self._same_padding(x)

    def _causal_padding(self, x):
        if self.kernel_size == 1:
            return x
        pad_l = self.kernel_size - 1
        pad_r = 0
        padding = [[0, 0], [0, 0], [pad_l, pad_r]]
        x = F.pad(x, commons.convert_pad_shape(padding))
        return x

    def _same_padding(self, x):
        if self.kernel_size == 1:
            return x
        pad_l = (self.kernel_size - 1) // 2
        pad_r = self.kernel_size // 2
        padding = [[0, 0], [0, 0], [pad_l, pad_r]]
        x = F.pad(x, commons.convert_pad_shape(padding))
        return x


class MRTE(nn.Module):
    def __init__(
        self,
        content_enc_channels=192,
        hidden_size=512,
        out_channels=192,
        kernel_size=5,
        n_heads=4,
        ge_layer=2,
    ):
        super(MRTE, self).__init__()
        self.cross_attention = MultiHeadAttention(hidden_size, hidden_size, n_heads)
        self.c_pre = nn.Conv1d(content_enc_channels, hidden_size, 1)
        self.text_pre = nn.Conv1d(content_enc_channels, hidden_size, 1)
        self.c_post = nn.Conv1d(hidden_size, out_channels, 1)

    def forward(self, ssl_enc, ssl_mask, text, text_mask, ge):
        attn_mask = text_mask.unsqueeze(2) * ssl_mask.unsqueeze(-1)

        ssl_enc = self.c_pre(ssl_enc * ssl_mask)
        text_enc = self.text_pre(text * text_mask)
        x = (
                self.cross_attention(
                    ssl_enc * ssl_mask, text_enc * text_mask, attn_mask
                )
                + ssl_enc
                + ge
            )
        x = self.c_post(x * ssl_mask)
        return x
